/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Debug printing facilities 
///////////////////////////////////////////////////////////////////////////////

#include <hw/debug.hpp>

namespace cpptesk {
namespace hw {

std::ostream& operator << (std::ostream& out, const debug_style_t &style)
{
    int i = 0;

    CPPTESK_ASSERTION(style.fg_color != DEFAULT && style.bg_color != DEFAULT,
        "Incorrect debug printing style");

    out << "\x1b[";

    if(style.fg_color == NORMAL || style.bg_color == NORMAL)
        { out << "0"; i++; }

    if(style.fg_color != NORMAL)
        { out << (i++ ? ";" : "") << std::dec << (30 + style.fg_color); }
    if(style.bg_color != NORMAL)
        { out << (i++ ? ";" : "") << std::dec << (40 + style.bg_color); }

    return out << "m";
}

// By default, all basic debug information is printed
debug_level_t Debug::_debug_level      = DEBUG_INFO;
debug_level_t Debug::_last_debug_level = DEBUG_NONE;

// By default, terminal highlighting is turned on
bool Debug::_debug_color = true;

unsigned Debug::_debug_count = 0;

unsigned Debug::_debug_indent = 0;

Debug::debug_config_t Debug::_debug_config[] =
{
    { "FAIL", { BLACK,  RED     }, { RED,     NORMAL }},
    { "WARN", { BLACK,  YELLOW  }, { YELLOW,  NORMAL }},
    { "USER", { NORMAL, NORMAL  }, { NORMAL,  NORMAL }},
    { "INFO", { BLACK,  GREEN   }, { NORMAL,  NORMAL }},
    { "MORE", { BLACK,  GREEN   }, { NORMAL,  NORMAL }},
    { "TOOL", { BLACK,  MAGENTA }, { NORMAL,  NORMAL }}
};

void Debug::setLevel(debug_level_t debug_level)
{
    _debug_level = debug_level;
}

void Debug::enableColoredPrint(bool debug_color)
{
    _debug_color = debug_color;
}

void Debug::setStyle(debug_level_t debug_level, debug_color_t fg_level, debug_color_t bg_level,
    debug_color_t fg_message, debug_color_t bg_message)
{
    debug_config_t &config = _debug_config[debug_level];

    debug_color_t fg_lvl = fg_level   == DEFAULT ? config.lvl_style.fg_color : fg_level;
    debug_color_t bg_lvl = bg_level   == DEFAULT ? config.lvl_style.bg_color : bg_level;
    debug_color_t fg_msg = fg_message == DEFAULT ? config.msg_style.fg_color : fg_message;
    debug_color_t bg_msg = bg_message == DEFAULT ? config.msg_style.bg_color : bg_message;

    _debug_config[debug_level].lvl_style = CPPTESK_DEBUG_STYLE(fg_lvl, bg_lvl);
    _debug_config[debug_level].msg_style = CPPTESK_DEBUG_STYLE(fg_msg, bg_msg);
}

unsigned Debug::setIndent(unsigned debug_indent)
{
    unsigned old = _debug_indent;

    _debug_indent = debug_indent;

    return old;
}

void Debug::incIndent()
{
    _debug_indent++;
}

void Debug::decIndent()
{
    CPPTESK_ASSERTION(_debug_indent > 0, "Cannot decrement debug indent");

    _debug_indent--;
}

void Debug::print(debug_level_t debug_level, const std::string &debug_message,
    debug_color_t fg_color, debug_color_t bg_color)
{
    std::ostream &out = CPPTESK_DEBUG_STREAM();
    debug_config_t &config = _debug_config[debug_level];

    if((int)_debug_level >= (int)debug_level)
    {
        if(debug_level != _last_debug_level)
        {
            // Separator
            out.width(80); out.fill('-');
            out << "" << std::endl;
        }

        if(_debug_color) { out << config.lvl_style; }

        // Debug message number
        out.width(6); out.fill('0');
        out << std::right << std::dec << _debug_count << " ";

        // Debug message tag
        out << "[" << config.tag << "]";

        if(_debug_color)
        {
            debug_color_t fg = (fg_color == DEFAULT ? config.msg_style.fg_color : fg_color);
            debug_color_t bg = (bg_color == DEFAULT ? config.msg_style.bg_color : bg_color);

            out << CPPTESK_DEBUG_STYLE(fg, bg);
        }

        // Debug indent
	out.width(4 * _debug_indent); out.fill(' ');
	// Debug message
	out << " " << debug_message;

        if(_debug_color) { out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL); }
        out << std::endl;

        _last_debug_level = debug_level;
        _debug_count = (_debug_count + 1) % 1000000;
    }
}

}} // namespace cpptesk::hw

