/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Model of a message
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/clone.hpp>
#include <hw/errors.hpp>
#include <hw/process.hpp>
#include <hw/random.hpp>
#include <hw/time.hpp>
#include <hw/unused.hpp>

#include <stdint.h>
#include <string.h>

#include <sstream>
#include <ostream>
#include <string>
#include <vector>

////////////////////////////////////////////////////////////////////////////////
// Declaration of Message Classes
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_MESSAGE
 * Declares a model message.
 */
#define CPPTESK_MESSAGE(UserMessage) \
    struct UserMessage: public ::cpptesk::hw::Message

#ifdef HW_SHORT_NAMES
    #define MESSAGE(UserMessage) CPPTESK_MESSAGE(UserMessage)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Declaration of Message Fields
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_DECLARE_MASKED_FIELD
 * Declares a masked field.
 */
#define CPPTESK_DECLARE_MASKED_FIELD(field_name, field_width, field_mask) \
    ::cpptesk::hw::Field<field_width, field_mask> field_name; \
    inline uint64_t get_##field_name() const { return field_name.get(); } \
    inline void set_##field_name(uint64_t value) { field_name.set(value); }

// IMPORTANT: To avoid warnings 1ull << 64 should be changed (1ull << 63) << 1.
#define __FIELD_MASK(field_width) \
    (((1ull << ((field_width) - 1)) << 1) - 1)

/** @def CPPTESK_DECLARE_FIELD
 * Declares a field.
 */
#define CPPTESK_DECLARE_FIELD(field_name, field_width) \
    CPPTESK_DECLARE_MASKED_FIELD(field_name, field_width, __FIELD_MASK(field_width))

/** @def CPPTESK_DECLARE_BIT
 * Declares a single-bit field.
 */
#define CPPTESK_DECLARE_BIT(field_name) \
    CPPTESK_DECLARE_FIELD(field_name, 1)

#ifdef HW_SHORT_NAMES
    #define DECLARE_MASKED_FIELD(field_name, field_width, field_mask) CPPTESK_DECLARE_MASKED_FIELD(field_name, field_width, field_mask)
    #define DECLARE_FIELD(field_name, field_width)                    CPPTESK_DECLARE_FIELD(field_name, field_width)
    #define DECLARE_BIT(field_name)                                   CPPTESK_DECLARE_BIT(field_name)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Declaration and Definition of a Comparator
////////////////////////////////////////////////////////////////////////////////

#if CPPTESKHW_CHECK_VERSION(110428)

    /** @def CPPTESK_DECLARE_COMPARATOR
     * Declares a comparator in a message class.
     */
    #define CPPTESK_DECLARE_COMPARATOR() \
        virtual std::string compare(const ::cpptesk::hw::Message &message) const

    /** @def CPPTESK_DEFINE_COMPARATOR
     * Defines a comparator.
     */
    #define CPPTESK_DEFINE_COMPARATOR(UserMessage) \
        std::string UserMessage::compare(const ::cpptesk::hw::Message &message) const

    #ifdef HW_SHORT_NAMES
        #define DECLARE_COMPARATOR() CPPTESK_DECLARE_COMPARATOR()
        #define DEFINE_COMPARATOR(UserMessage)  CPPTESK_DEFINE_COMPARATOR(UserMessage)
    #endif // HW_SHORT_NAMES

#else

    /** @def CPPTESK_DECLARE_COMPARATOR
     * Declares a comparator in a message class.
     */
    #define CPPTESK_DECLARE_COMPARATOR(UserMessage) \
        /* Overloads the basic comparator */ \
        virtual std::string compare(const ::cpptesk::hw::Message &message) const \
            { return compare(dynamic_cast<const UserMessage&>(message)); } \
        virtual std::string compare(const UserMessage &message) const

    /** @def CPPTESK_DEFINE_COMPARATOR
     * Defines a comparator.
     */
    #define CPPTESK_DEFINE_COMPARATOR(UserMessage) \
        std::string UserMessage::compare(const UserMessage &message) const

    #ifdef HW_SHORT_NAMES
        #define DECLARE_COMPARATOR(UserMessage) CPPTESK_DECLARE_COMPARATOR(UserMessage)
        #define DEFINE_COMPARATOR(UserMessage)  CPPTESK_DEFINE_COMPARATOR(UserMessage)
    #endif // HW_SHORT_NAMES

#endif // CPPTESKHW_CHECK_VERSION

/** @def CPPTESK_DECLARE_RANDOMIZER
 * Declares a randomizer in a message class.
 */
#define CPPTESK_DECLARE_RANDOMIZER() \
    virtual void randomize()

/** @def CPPTESK_DEFINE_RANDOMIZER
 * Defines a randomizer.
 */
#define CPPTESK_DEFINE_RANDOMIZER(UserMessage) \
    void UserMessage::randomize()

#ifdef HW_SHORT_NAMES
    #define DECLARE_RANDOMIZER()            CPPTESK_DECLARE_RANDOMIZER()
    #define DEFINE_RANDOMIZER(UserMessage)  CPPTESK_DEFINE_RANDOMIZER(UserMessage)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Requirements Traceability
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_SET_REQUIREMENT
 * Sets a requirement (some string explaning a reason for sending or receiving a message).
 */
#define CPPTESK_SET_REQUIREMENT(msg, req) \
    _cpptesk_set_explanation(msg, time(), time() + getReactionTimeout(), __FILE__, __LINE__, req)

/** @def CPPTESK_SET_REQUIREMENT
 * Gets a message's requirement.
 */
#define CPPTESK_GET_REQUIREMENT(msg) \
    ((msg).getExplanation())

/** @def CPPTESK_SET_COMMENT
 * Alias of CPPTESK_SET_REQUIREMENT
 */
#define CPPTESK_SET_COMMENT(msg, req) CPPTESK_SET_REQUIREMENT(msg, req)

/** @def CPPTESK_GET_COMMENT
 * Alias of CPPTESK_GET_REQUIREMENT
 */
#define CPPTESK_GET_COMMENT(msg) CPPTESK_GET_REQUIREMENT(msg)

#ifdef HW_SHORT_NAMES
    #define SET_REQUIREMENT(msg, req) CPPTESK_SET_REQUIREMENT(msg, req)
    #define GET_REQUIREMENT(msg)      CPPTESK_GET_REQUIREMENT(msg)
    #define SET_COMMENT(msg, req)     CPPTESK_SET_COMMENT(msg, req)
    #define GET_COMMENT(msg)          CPPTESK_GET_COMMENT(msg)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Field Registration
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_ADD_FIELD
 * Registers a field in a message class.
 */
#define CPPTESK_ADD_FIELD(full_field_name) \
    registerField(#full_field_name, (Message::field_t)&full_field_name, true)

#define CPPTESK_ADD_INCOMPARABLE_FIELD(full_field_name) \
    registerField(#full_field_name, (Message::field_t)&full_field_name, false)

#ifdef HW_SHORT_NAMES
    #define ADD_FIELD(full_field_name)              CPPTESK_ADD_FIELD(full_field_name)
    #define ADD_INCOMPARABLE_FIELD(full_field_name) CPPTESK_ADD_INCOMPARABLE_FIELD(full_field_name)
#endif // HW_SHORT_NAMES

/////////////////////////////////////////////////////////////////////////////////
// Clone Support
/////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_SUPPORT_CLONE(UserMessage) \
    virtual UserMessage* clone() const { return new UserMessage(*this); }

#ifdef HW_SHORT_NAMES
    #define SUPPORT_CLONE(UserMessage) CPPTESK_SUPPORT_CLONE(UserMessage)
#endif // HW_SHORT_NAMES

/////////////////////////////////////////////////////////////////////////////////
// Message Transmission Delay & Priority
/////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_MESSAGE_DELAY(msg, delay) \
    (msg).setDelay(delay)

#define CPPTESK_MESSAGE_LOCK(msg, pid) \
    (msg).setLock(pid)

#ifdef HW_SHORT_NAMES
    #define MESSAGE_DELAY(msg, delay) CPPTESK_MESSAGE_DELAY(msg, delay)
    #define MESSAGE_LOCK(msg, pid)    CPPTESK_MESSAGE_LOCK(msg, pid)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Message randomization
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_RANDOMIZE_MESSAGE
 * Randomizes a message object.
 */
#define CPPTESK_RANDOMIZE_MESSAGE(msg) \
    (msg).randomize()

#ifdef HW_SHORT_NAMES
    #define RANDOMIZE_MESSAGE(msg) CPPTESK_RANDOMIZE_MESSAGE(msg)
#endif // HW_SHORT_NAMES

#define __CPPTESK_CLONE_MESSAGE(msg) \
    ((msg) ? (msg)->clone() : 0)

namespace cpptesk {
namespace hw {

inline uint16_t reverse_byte_order(uint16_t value)
{
    return ((value & 0xff) << 8)
         | ((value >> 8) & 0xff);
}

inline uint32_t reverse_byte_order(uint32_t value)
{
    return (uint32_t)reverse_byte_order((uint16_t)((value & 0xffff))) << 16
         | (uint32_t)reverse_byte_order((uint16_t)((value >> 16) & 0xffff));
}

inline uint64_t reverse_byte_order(uint64_t value)
{
    return (uint64_t)reverse_byte_order((uint32_t)((value & 0xffffFFFFul))) << 32
         | (uint64_t)reverse_byte_order((uint32_t)((value >> 32) & 0xffffFFFFul));
}

inline uint64_t get_mask(unsigned width)
{
    return width == 64 ? 0xffffFFFFffffFFFFull : ((1ull << width) - 1);
}

inline uint64_t get_mask(unsigned min, unsigned max)
{
    return get_mask((max - min) + 1) << min;
}

inline uint64_t get_field(uint64_t value, unsigned min, unsigned max)
{
    return (value >> min) & get_mask((max - min) + 1);
}

inline uint64_t get_field(uint64_t value, unsigned bit)
{
    return get_field(value, bit, bit);
}

inline void set_field(uint64_t &value, unsigned min, unsigned max, uint64_t field)
{
    uint64_t mask = get_mask(min, max);
    value &= ~mask; value |= (field << min) & mask;
}

inline void set_field(uint64_t &value, unsigned bit, uint64_t field)
{
    set_field(value, bit, bit, field);
}

typedef enum interlock_t
{
    LOCK_NONE       = 0x00,
    LOCK_HOLD       = 0x01,
    LOCK_DELAY      = 0x02,
    LOCK_HOLD_DELAY = LOCK_HOLD | LOCK_DELAY
} interlock_t;

////////////////////////////////////////////////////////////////////////////////
// Message Field
////////////////////////////////////////////////////////////////////////////////

template<size_t field_width = 64, uint64_t field_mask = 0xffffFFFFffffFFFFull>
class Field
{
public:
    Field(uint64_t value = 0):
        _short_name("<unknown>"), _base_name("<unknown>"), _full_name("<unknown>"),
        _data(value), _compare(true), _valid(false) {}

    Field(const Field &rhs) { _copy(rhs); }

    virtual ~Field() {}

    operator uint64_t () const { return get(); }

    Field& operator  = (const Field &rhs) { set(rhs.get());         return *this; }
    Field& operator |= (const Field &rhs) { set(get() | rhs.get()); return *this; }
    Field& operator &= (const Field &rhs) { set(get() & rhs.get()); return *this; }
    Field& operator ^= (const Field &rhs) { set(get() ^ rhs.get()); return *this; }

    inline uint64_t get() const { return _data; }

    inline void set(uint64_t value)
    {
        CPPTESK_ASSERTION((_data & ~get_mask(field_width)) == 0, "Cannot assign " << std::hex << value
            << " to " << getFullName() << "[" << std::dec << field_width << "]");

        _data  = value & get_mask(field_width) & field_mask;
        _valid = true;
    }

    inline std::string getShortName() const { return _short_name; }
    inline std::string getBaseName()  const { return _base_name; }
    inline std::string getFullName()  const { return _full_name; }

    inline void parseFullName(const std::string &full_name)
    {
        size_t i;

        setFullName(full_name);

        if((i = full_name.find("::")) == std::string::npos)
        {
            setBaseName("");
            setShortName(full_name);
        }
        else
        {
            setBaseName(full_name.substr(0, i));
            setShortName(full_name.substr(i + 2));
        }
    }

    inline bool isComparable() const { return _compare; }
    inline void setComparable(bool compare = true) { _compare = compare; }

    inline bool isValid() const { return _valid; }
    inline void setValid(bool valid = true) { _valid = valid; }
    inline void invalidate() { _valid = false; }

    virtual std::string compare(const Field &rhs) const
    {
        std::stringstream buffer;

        if(!(isValid() && rhs.isValid()))
        {
            CPPTESK_DEBUG_PRINT(DEBUG_WARN, getBaseName() << (!isValid() ? "{spec=" : "{impl=")
                << toString() << "} is not initialized");
	    return "";
        }

        // IMPORTANT: When calling compare() the fields are always compared.
	// For incomparable fields compare() is not called by the message.
        if(get() != rhs.get())
            { buffer << getBaseName() << "{spec=" << toString() << ", " << "impl=" << rhs.toString() << "}"; }

        return buffer.str();
    }

    virtual std::string toString() const
    {
        std::stringstream buffer;

        if(isValid())
            { buffer << getShortName() << "(" << std::hex << get() << ")"; }
	else
            { buffer << getShortName() << "(<unknown>)"; }

        return buffer.str();
    }

    virtual void randomize() { set(Random::create().rand<uint64_t>(field_width)); }

protected:
    inline void setShortName(const std::string &short_name) { _short_name = short_name; }
    inline void setBaseName(const std::string &base_name)   { _base_name  = base_name; }
    inline void setFullName(const std::string &full_name)   { _full_name  = full_name; }

protected:
    void _copy(const Field &rhs)
    {
        _short_name = rhs._short_name;
        _base_name  = rhs._base_name;
        _full_name  = rhs._full_name;
        _data       = rhs._data;
        _compare    = rhs._compare;
        _valid      = rhs._valid;
    }

protected:
    /// Field's short name.
    std::string _short_name;
    /// Message's name
    std::string _base_name;
    /// Field's full name.
    std::string _full_name;
    /// Field's value.
    uint64_t _data;
    /// Compare or not.
    bool _compare;
    /// Validity.
    bool _valid;
};

class Message:
    public Randomizable
  , public Cloneable
{
public:
    /// Successfull comparison return value (empty string).
    const static std::string COMPARE_OK;

public:
    /// Comparison flags
    enum CompareFlags { FIRST_MISMATCH_ONLY = 0x01 };

    Message(): _valid(true), _explanation("<unknown>"), _compare_flags(0), _detected(false)
        { memset(&_sysdata, 0, sizeof(_sysdata)); }

    virtual ~Message() {}

    Message(const Message &rhs);

    Message& operator = (const Message &rhs);

#ifdef CPPTESK_CLONE_INPUT_MESSAGE
    virtual Message* clone() const = 0;
#else
    virtual Message* clone() const { return this; }
#endif

    std::string getName() const { return _name; }
    void setName(const std::string &name) { _name = name; }

    void setSystemData(priority_t priority, interlock_t interlock, pid_t pid, cycle_t delay = 0)
    {
        _sysdata.priority  = priority;
        _sysdata.interlock = interlock;
        _sysdata.pid       = pid;
        _sysdata.delay     = delay;
    }

    void setPriority(priority_t priority) { _sysdata.priority = priority; }
    void setDelay(cycle_t delay)          { _sysdata.interlock = LOCK_DELAY; _sysdata.delay = delay; }
    void setHold(pid_t pid)               { _sysdata.interlock = LOCK_HOLD; _sysdata.pid = pid; }

    void setExplanation(const std::string &explanation) { _explanation = explanation; }
    const std::string& getExplanation() const { return _explanation; }

    priority_t getPriority() const { return _sysdata.priority; }
    interlock_t getInterlock() const { return _sysdata.interlock; }
    pid_t getPID() const { return _sysdata.pid; }
    int getDelay() const { return _sysdata.delay; }

    void setValid(bool valid = true) { _valid = valid; }
    void validate() { _valid = true; }
    void invalidate() { _valid = false; }
    bool isValid() const { return _valid; }

    void setValidFields(bool valid = true);
    void validateFields();
    void invalidateFields();

    bool isDetected() const { return _detected; }
    void setDetected(bool detected = true) { _detected = detected; }
    
    void setCompareFlags(int flags) { _compare_flags = flags; }

    /// Compares a message with another one and returns non-empty string if they are not equal.
    virtual std::string compare(const Message &message) const;

    /// Returns a string representation of a message.
    virtual std::string toString() const;
    /// Randomizes a message data.
    virtual void randomize();

protected:
    /// Registry of message fields.
    typedef Field<> Message::*field_t;
    typedef std::vector<field_t> fields_list_t;

    void registerField(const std::string &full_name, field_t field, bool compare = true)
    {
        (this->*field).parseFullName(full_name);
        (this->*field).setComparable(compare);

        _fields.push_back(field);

        setName((this->*field).getBaseName());
    }

    void copyFields(const Message &rhs);

protected:
    std::string _name;
    fields_list_t _fields;
    bool _valid;
    std::string _explanation;

    struct
    {
        priority_t priority;
        interlock_t interlock;
        pid_t pid;
        cycle_t delay;
    } _sysdata;

    int _compare_flags;
    bool _detected;
};

/**
 * Empty message being used as an input message for input-independent processes,
 * like reaction listeners and other infinite loops.
 */
CPPTESK_MESSAGE(EmptyMessage)
{
    EmptyMessage(): Message() {}
    EmptyMessage(const EmptyMessage &rhs): Message(rhs) {}

    virtual ~EmptyMessage() {}

    CPPTESK_SUPPORT_CLONE(EmptyMessage)
};

extern EmptyMessage CPPTESK_DEFAULT_MESSAGE;

inline std::ostream& operator << (std::ostream& out, const Message &msg)
{
    return out << msg.toString();
}

// Internal function that is used in some defines.
inline Message& _cpptesk_set_explanation(Message &msg, cycle_t time_min, cycle_t time_max,
    const std::string &file, int line, const std::string &req = "")
{
    std::stringstream buffer;

    buffer << req << (!req.empty() ? " " : "")  << "cycle=["
           << std::dec << time_min << "," << time_max << "]";
#if 0
    // The information on file and line is kept in the proces call descriptor
    buffer << " (" << file << ":" << line << ")";
#else
    CPPTESK_UNUSED_ARG(file);
    CPPTESK_UNUSED_ARG(line);
#endif

    msg.setExplanation(buffer.str());

    return msg;
}

template<class UserMessage>
inline UserMessage& _cpptesk_cast_message(Message &message, const std::string &file, int line)
{
    UserMessage *casted_message = dynamic_cast<UserMessage*>(&message);
    __CPPTESK_ASSERTION_FILE_LINE(casted_message, "Cannot not cast a message to the type given", file, line);
    return *casted_message;
}

template<class UserMessage>
inline const UserMessage& _cpptesk_cast_message(const Message &message, const std::string &file, int line)
{
    const UserMessage *casted_message = dynamic_cast<const UserMessage*>(&message);
    __CPPTESK_ASSERTION_FILE_LINE(casted_message, "Cannot not cast a message to the type given", file, line);
    return *casted_message;
}

template<class UserMessage>
inline const UserMessage& _cpptesk_const_cast_message(Message &message, const std::string &file, int line)
{
    const UserMessage *casted_message = dynamic_cast<const UserMessage*>(&message);
    __CPPTESK_ASSERTION_FILE_LINE(casted_message, "Cannot not cast a message to the type given", file, line);
    return *casted_message;
}

template<class UserMessage>
inline const UserMessage& _cpptesk_const_cast_message(const Message &message, const std::string &file, int line)
{
    const UserMessage *casted_message = dynamic_cast<const UserMessage*>(&message);
    __CPPTESK_ASSERTION_FILE_LINE(casted_message, "Cannot not cast a message to the type given", file, line);
    return *casted_message;
}

}} // namespace cpptesk::hw
