/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Model of an operation execution process
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/time.hpp>

#include <stdint.h>
#include <ucontext.h>

#include <iostream>
#include <sstream>
#include <list>
#include <string>

namespace cpptesk {
namespace hw {

class Interface;
class Message;

#define CPPTESK_INCORRECT_PID (-1)

#ifdef HW_SHORT_NAMES
    #define INCORRECT_PID CPPTESK_INCORRECT_PID
#endif // HW_SHORT_NAMES

/** @typedef pid_t
 * Process identifier.
 */
typedef int pid_t;

#define __CPPTESK_DEFINE_PRIORITY(name, value) \
    PRIORITY_##name = LOWEST_PRIORITY + value

#define __CPPTESK_DEFINE_10_PRIORITIES(prefix, value) \
    __CPPTESK_DEFINE_PRIORITY(prefix##0, value + 0), \
    __CPPTESK_DEFINE_PRIORITY(prefix##1, value + 1), \
    __CPPTESK_DEFINE_PRIORITY(prefix##2, value + 2), \
    __CPPTESK_DEFINE_PRIORITY(prefix##3, value + 3), \
    __CPPTESK_DEFINE_PRIORITY(prefix##4, value + 4), \
    __CPPTESK_DEFINE_PRIORITY(prefix##5, value + 5), \
    __CPPTESK_DEFINE_PRIORITY(prefix##6, value + 6), \
    __CPPTESK_DEFINE_PRIORITY(prefix##7, value + 7), \
    __CPPTESK_DEFINE_PRIORITY(prefix##8, value + 8), \
    __CPPTESK_DEFINE_PRIORITY(prefix##9, value + 9)

#define __CPPTESK_DEFINE_100_PRIORITIES(prefix, value) \
    __CPPTESK_DEFINE_10_PRIORITIES(prefix##0, value + 00), \
    __CPPTESK_DEFINE_10_PRIORITIES(prefix##1, value + 10), \
    __CPPTESK_DEFINE_10_PRIORITIES(prefix##2, value + 20), \
    __CPPTESK_DEFINE_10_PRIORITIES(prefix##3, value + 30), \
    __CPPTESK_DEFINE_10_PRIORITIES(prefix##4, value + 40), \
    __CPPTESK_DEFINE_10_PRIORITIES(prefix##5, value + 50), \
    __CPPTESK_DEFINE_10_PRIORITIES(prefix##6, value + 60), \
    __CPPTESK_DEFINE_10_PRIORITIES(prefix##7, value + 70), \
    __CPPTESK_DEFINE_10_PRIORITIES(prefix##8, value + 80), \
    __CPPTESK_DEFINE_10_PRIORITIES(prefix##9, value + 90)

/** @typedef priority_t
 * Using null priority is prohibited.
 */
typedef enum priority_t
{
    /// Reserved value that is equal NORMAL_PRIORITY
    NULL_PRIORITY    = 0x00,

    LOWEST_PRIORITY  = 0x01,
    NORMAL_PRIORITY  = 0x10,
    HIGHEST_PRIORITY = 0xff,

    /// The following things are satisfied:
    /// LOWEST_PRIORITY < NORMAL_PRIORITY <  HIGHEST_PRIORITY
    /// LOWEST_PRIORITY < PRIORITY_01 < ... < PRIORITY_99 < HIGHEST_PRIORITY

    /// PRIORITY_00, ..., PRIORITY_99
    __CPPTESK_DEFINE_100_PRIORITIES(,  000),
    /// PRIORITY_000, ..., PRIORITY_099
    __CPPTESK_DEFINE_100_PRIORITIES(0, 000),
    /// PRIORITY_100, ..., PRIORITY_199
    __CPPTESK_DEFINE_100_PRIORITIES(1, 100)
} priority_t;

#define __CPPTESK_PRIORITY_MASK 0xff

/** @typedef status_t
 * Process status.
 */
typedef enum status_t
{
    /// Process is still executing.
    EXECUTING = -1,
    /// Process execution has been successfully completed.
    COMPLETED =  0,
    /// Precondition violation.
    UNPREDICT =  1,
    /// Postcondition violation.
    VIOLATION =  2,
} status_t;

/** @typedef point_t
 * Synchronization point type.
 */
typedef enum point_t
{
    /// Initial synchronization point (process is not executed).
    START = 0,
    /// Synchronization caused by CYCLE() or DELAY().
    CYCLE = 1,
    /// Synchronization caused by GUARD(condition) or other wait condition.
    GUARD = 2,
} point_t;

class Process
{
public:
    /// Process identifier.
    typedef pid_t Id;

    /// Descriptor of a point in code in which a process is called.
    struct CallDescriptor
    {
        CallDescriptor() { reset(); }

        CallDescriptor(const std::string &name, const std::string &file = "<unknown>", int line = -1):
            name(name), file(file), line(line) {}

        void reset()
        {
            name = "<unknown>";
            file = "<unknown>";
            line = -1;
        }

        /// Name of a process function.
        std::string name;
        /// File name where a process is called.
        std::string file;
        /// Line in the file where a process is called.
        int line;
    };

    /// Descriptor of a synchronization point within a process.
    struct PointDescriptor
    {
        PointDescriptor(point_t type = START, const std::string &name = "START()", const std::string &file = "<unknown>", int line = -1):
            type(type), name(name), file(file), line(line), timer(0) {}

        /// Point type.
        point_t type;
 	/// Point name.
	std::string name;
       /// File name where a point is defined.
        std::string file;
        /// Line in the file where a point is defined.
        int line;
        /// Point execution time (for points of the WAIT type).
        cycle_t timer;
    };

    /// Call stack.
    typedef std::list<CallDescriptor> call_stack_t;

public:
    Process(pid_t pid, Interface &iface, Message &message, bool clone_message = true);
    virtual ~Process();

    pid_t getPID() const { return pid; }

    priority_t getPriority() const { return this->priority; }
    void setPriority(priority_t priority) { this->priority = priority; }

    bool isStarted() const { return point.type != START; }
    bool isWaiting() const { return point.type == GUARD; }
    bool isActive()  const { return point.type == CYCLE; }

    void cycle() { clk_total++; point.timer++; }

    ////////////////////////////////////////////////////////////////////////////
    // For internal use only
    ////////////////////////////////////////////////////////////////////////////

    /// Process identifier.
    pid_t pid;
    /// Process priority.
    priority_t priority;
     /// Status of the process execution.
    status_t status;
    /// Process context.
    ucontext_t* ucp;
    /// Starting time.
    cycle_t clk_start;
    /// Process execution time (number of clock cycles passed from the process start-up until now).
    cycle_t clk_total;
    /// Timeout.
    bool timeout;
    /// Input interface.
    Interface *input_iface;
    /// Copy of an input message.
    Message *input_message;
    /// Original message.
    Message *original_message;

    ////////////////////////////////////////////////////////////////////////////
    // Current Synchronization Point Descriptor
    ////////////////////////////////////////////////////////////////////////////
    
    /// Current point.
    PointDescriptor point;

    ////////////////////////////////////////////////////////////////////////////
    // Call Descriptors and Call Stack
    ////////////////////////////////////////////////////////////////////////////

    /// Prepares a call descriptor of a process.
    void prepare_call(const CallDescriptor &call)
    {
        if(call_stack.empty())
            { parallel_call = call; }

        sequential_call = call;
    }

    /// Returns size of the call stack.
    size_t call_stack_size() const { return call_stack.size();  }

    /// Returns size of the call stack.
    const CallDescriptor& call_stack_top()    const { return call_stack.back();   }
    const CallDescriptor& call_stack_bottom() const { return call_stack.front();  }

    /// Adds a call descriptor into the stack.
    void start_call()
        { call_stack.push_back(sequential_call); sequential_call.reset(); }

    /// Removes a call descriptor from the stack. Returns true iff it's the last item.
    void stop_call()
        { call_stack.pop_back(); }

    /// Descriptor of the process call (bottom of the call stack).
    CallDescriptor parallel_call;
    /// Descriptor of a currently called sequential process within the process.
    CallDescriptor sequential_call;

    /// Stack of sequential calls within the process (contains this process's call in the bottom).
    call_stack_t call_stack;
};

bool greaterProcessPriority(Process *lhs, Process *rhs);
bool lessProcessPriority(Process *lhs, Process *rhs);

inline std::ostream& operator << (std::ostream& out, const Process::CallDescriptor &call)
{
    // Parameters are unsupported.
    out << call.name << "()";

    if(call.line >= 0)
        { out << " at " << call.file << ":" << std::dec << call.line; }

    return out;
}

inline std::ostream& operator << (std::ostream& out, const Process::call_stack_t &call_stack)
{
    Process::call_stack_t::const_reverse_iterator i;

    // IMPORTANT: Call stack can be empty before START()
    if(call_stack.empty()) { return out; }

    out << " in " << *(i = call_stack.rbegin());

    for(++i; i != call_stack.rend(); i++)
        { out << std::endl << " in " << *i; }

    return out;
}

inline std::ostream& operator << (std::ostream& out, const Process &process)
{
    // IMPORTANT: Call stack can be empty before START()
    const Process::CallDescriptor &call = (process.call_stack_size() > 0 ? process.call_stack_top() : process.parallel_call);

    return out << call.name << "[" << process.pid << "]";
}

}} // namespace cpptesk::hw

