/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Sequence
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>

#include <hw/sequence/delay.hpp> 
#include <hw/sequence/event.hpp>
#include <hw/sequence/guard.hpp>
#include <hw/sequence/predicate.hpp>
#include <hw/sequence/repeat.hpp>

#include <hw/utils/shared_ptr.hpp>

#include <list>
#include <iostream>

namespace cpptesk {
namespace hw {
namespace sequence {

////////////////////////////////////////////////////////////////////////////////
// Sequence
////////////////////////////////////////////////////////////////////////////////

/**
 * Sequence syntax.
 */
class Sequence
{
public:
    /// Sequence type.
    enum type_t
    {
        /// A single event.
        EVENT,
        /// A guard condition.
        GUARD,
        /// A sequence (empty or non-empty).
        SEQ,
        /// AND-composition of sequences.
        AND,
        /// OR-composition of sequences.
        OR,
        /// Repetition of a sequence.
        REPEAT
    };

    typedef std::list<Sequence>::iterator iterator;
    typedef std::list<Sequence>::const_iterator const_iterator;

public:

    ////////////////////////////////////////////////////////////////////////////
    // Default constructor
    ////////////////////////////////////////////////////////////////////////////

    Sequence(): type(SEQ)
    {
        setDefaultValues();
    }

    Sequence(const Event &event, type_t type): type(type), event(event.clone())
    {
        setDefaultValues();
    }

    Sequence(const Sequence &lhs, const Sequence &rhs, type_t type): type(type)
    {
        setDefaultValues();

        children.push_back(lhs);
        children.push_back(rhs);
    }

    Sequence(const Sequence &seq, const Repeat &rep): type(REPEAT)
    {
        setDefaultValues();

        children.push_back(seq);
        repetition.set(rep);
    }

    ////////////////////////////////////////////////////////////////////////////
    // Type casting constructors
    ////////////////////////////////////////////////////////////////////////////

    Sequence(const Delay &delay): type(SEQ)
    {
        setDefaultValues(delay);
    }

    Sequence(unsigned delay): type(SEQ)
    {
        setDefaultValues(delay);
    }

    Sequence(const Event &event): type(EVENT), event(event.clone())
    {
        setDefaultValues();
    }

    Sequence(const Guard &guard): type(GUARD), event(guard.getEvent().clone())
    {
        setDefaultValues();
    }

    virtual ~Sequence() {}

    /// Returns the sequence type.
    type_t getType()  const { return type; }

    bool isEvent()    const { return type == EVENT; }
    bool isGuard()    const { return type == GUARD; }
    bool isEmpty()    const { return type == SEQ &&  children.empty(); }
    bool isSequence() const { return type == SEQ && !children.empty(); }
    bool isAnd()      const { return type == AND; }
    bool isOr()       const { return type == OR;  }
    bool isRepeat()   const { return type == REPEAT; }

    /// Returns the delay before the sequence.
    const Delay& getDelay() const { return delay; }
    /// Returns true iff the delay before the sequence is set explicitly.
    bool isExplicitDelay() const { return explicit_delay; }

    /// Returns the event (for EVENT and GUARD only).
    const Event& getEvent() const { CPPTESK_ASSERTION(!event.isNull(), "Internal error: event is null"); return *event; }

    /// Returns the number of repetition of the sequence (for REPEAT sequences only).
    const Repeat& getRepeat() const { return repetition; }
    /// Returns true iff the sequence is repeated exactly one time.
    bool isSingleRepeated() const { return !isRepeat() || repetition.isOne(); }

    /// Returns the time constraints on sequence recognition.
    const Delay& getTime() const { return time; }
    /// Sets the time constraints on sequence recognition.
    void setTime(unsigned min, unsigned max) { time.set(min, max); }

    /// Returns the lhs child.
    const Sequence& getLhsChild() const { return children.front(); }
    /// Returns the rhs child.
    const Sequence& getRhsChild() const { return children.back(); }
    /// Returns the child.
    const Sequence& getChild()    const { return children.front(); }

    // Returns the iterator pointing to the first child.
    iterator begin() { return children.begin(); }
    const_iterator begin() const { return children.begin(); }
    // Returns the iterator indicating the end.
    iterator end() { return children.end(); }
    const_iterator end() const { return children.end(); }

    ////////////////////////////////////////////////////////////////////////////////
    // Concatenation of sequences
    ////////////////////////////////////////////////////////////////////////////////

    // IMPORTANT: Concatenation is implemented as an external operator <<

    ////////////////////////////////////////////////////////////////////////////////
    // AND and OR composition of sequences
    ////////////////////////////////////////////////////////////////////////////////

    Sequence operator && (const Sequence &rhs) const { return Sequence(*this, rhs, Sequence::AND); }
    Sequence operator &  (const Sequence &rhs) const { return Sequence(*this, rhs, Sequence::AND); }
    Sequence operator || (const Sequence &rhs) const { return Sequence(*this, rhs, Sequence::OR);  }
    Sequence operator |  (const Sequence &rhs) const { return Sequence(*this, rhs, Sequence::OR);  }

    ////////////////////////////////////////////////////////////////////////////
    // Repetition of a sequence
    ////////////////////////////////////////////////////////////////////////////

    Sequence repeat(unsigned m, unsigned n) const;

    Sequence repeat(unsigned n)   const { return repeat(n, n);           }
    Sequence no_repeat()          const { return repeat(1);              }
    Sequence any_repeat()         const { return repeat(0, Repeat::INF); }
    Sequence at_least(unsigned n) const { return repeat(n, Repeat::INF); }
    Sequence at_most(unsigned n)  const { return repeat(0, n);           }
    Sequence repeat()             const { return repeat(1, Repeat::INF); }

    Sequence operator () (unsigned m, unsigned n) const { return repeat(m, n); }
    Sequence operator () (unsigned n)             const { return repeat(n);    }

    ////////////////////////////////////////////////////////////////////////////////
    // Guard condition
    ////////////////////////////////////////////////////////////////////////////////

protected:

    /// Sets default values to the delay and repetition properties.
    void setDefaultValues()
    {
        delay.set(0);
        // The delay is not set explicitly
        explicit_delay = false;
        repetition.set(1);
        time.set(0, Delay::INF);
    }

    /// Sets default values to the delay and repetition properties.
    void setDefaultValues(Delay value)
    {
        delay.set(value);
        // The delay is set explicitly
        explicit_delay = true;
        repetition.set(1);
        time.set(0, Delay::INF);
    }

protected:

    friend std::ostream& operator << (std::ostream &out, const Sequence &seq);
    friend Sequence operator << (const Sequence &lhs, const Sequence &rhs);

    /// Sequence type.
    type_t type;

    /// Pointer to an event (for EVENT and GUARD sequences only).
    SharedPtr<Event> event;

    /// Child subsequences:
    /// (*) Empty list for EVENT and GUARD;
    /// (*) No restrictions for SEQ;
    /// (*) List of two subsequences for AND and OR;
    /// (*) List of one subsequence for REPEAT.
    std::list<Sequence> children; 

    /// Delay right before the sequence (the default value is [0,0]).
    /// For SEQ, the delay is ignored (it's taken from the first item).
    Delay delay;

    /// Indicates that the delay before the sequence is set explicitly
    /// (if it's not set explicitly, the default value is [1,1]).
    bool explicit_delay;

    /// Number of repetition of the sequence (the default value is [1,1]) (for REPEAT sequences only).
    Repeat repetition;

    /// Restriction on sequence recongnition time (the default value is [0,inf)).
    /// It makes sense for the top-level sequences only.
    Delay time;
};

////////////////////////////////////////////////////////////////////////////////
// Concatenation of sequences
////////////////////////////////////////////////////////////////////////////////

Sequence operator << (const Sequence &lhs, const Sequence &rhs);

////////////////////////////////////////////////////////////////////////////////
// Repetition of a sequence
////////////////////////////////////////////////////////////////////////////////

/// Temporal operator repeat(S, m, n) repeats a sequence S from m to n times.
/// Example: repeat(E1 << any_delay() << E2, 1, 2);
inline Sequence repeat(const Sequence &seq, unsigned m, unsigned n)
{
    return seq.repeat(m, n);
}

/// Temporal operator repeat(S, n) repeats a sequence S exactly n times.
/// Example: repeat(E1 << no_delay() << E2, 100);
inline Sequence repeat(const Sequence &seq, unsigned n)
{
    return seq.repeat(n);
}

/// Temporal operator no_repeat(S) is equivalent to repeat(S, 1).
/// Example: no_repeat(E1 << delay(1) << E2);
inline Sequence no_repeat(const Sequence &seq)
{
    return seq.no_repeat();
}

/// Temporal operator any_repeat(S) is equivalent to repeat(S, 0, inf).
/// Example: any_repeat(E1 << 10 << E2);
inline Sequence any_repeat(const Sequence &seq)
{
    return seq.any_repeat();
}

/// Temporal operator at_least(S, n) is equivalent to repeat(S, n, inf).
/// Example: at_least(E1 << no_delay() << E2, 10);
inline Sequence at_least(const Sequence &seq, unsigned n)
{
    return seq.at_least(n);
}

/// Temporal operator at_most(S, n) is equivalent to repeat(S, 0, n).
/// Example: at_most(E1 << 666 << E2, 0);
inline Sequence at_most(const Sequence &seq, unsigned n)
{
    return seq.at_most(n);
}

/// Temporal operator repeat(S) is equivalent to repeat(S, 1, inf).
/// Example: repeat(E1 << 1 << E2);
inline Sequence repeat(const Sequence &seq)
{
    return seq.repeat();
}

////////////////////////////////////////////////////////////////////////////////
// Sequence printing operator
////////////////////////////////////////////////////////////////////////////////

std::ostream& operator << (std::ostream &out, const Sequence &seq);

}}} // namespace cpptesk::hw::sequence

