/*
 * Copyright 2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * File: aspect_coverage.c
 * Description:
 *       This file contains implementation of Aspectrace aspect "coverage".
 */

#include <stdio.h>

#include <utils/assertion.h>
#include "aspect_coverage.h"
#include "aspectrace_internals.h"

static const char ASPECT[] = "coverage";


struct aspect_coverage_element  {
	const char *name;
	struct aspect_coverage_element *firstChild;
	struct aspect_coverage_element *nextSibling;
};

aspect_coverage_element *create_CoverageElement(const char *name, aspect_coverage_element **children, size_t count)
{
aspect_coverage_element *res;
aspect_coverage_element *cur;

	assertion( name != NULL, FORMAT("create_CoverageElement: NULL name") );

    res = malloc( sizeof(aspect_coverage_element) );
	assertion( res != NULL, FORMAT("create_CoverageElement: No memory") );
#ifdef ALLOC_NEW
	res->name = strdup(name);
#else
	res->name = name;
#endif

    res->firstChild = NULL;
	res->nextSibling = NULL;
	if (children != NULL && count > 0)
	{
	    size_t ctr;
		res->firstChild = cur = children[0];
		for (ctr = 0; ctr < count-1; ctr++) {
			cur->nextSibling = children[ ctr+1 ];
			cur = cur->nextSibling;
		}
		cur->nextSibling = NULL;
	}
	return res;
}

aspect_coverage_element *create_CoverageElement_a(const char *name, aspect_coverage_element **children)
{
    return create_CoverageElement(name, children, list_len((const void**)children));
}

/** @return next sibling */
static aspect_coverage_element * destroy_CoverageElement(aspect_coverage_element *elem)
{
aspect_coverage_element *next;

#ifdef ALLOC_NEW
	free( elem->name );
#endif

	while (elem->firstChild != NULL) {
		elem->firstChild = destroy_CoverageElement( elem->firstChild );
	}

	next = elem->nextSibling;  /* will be cleared by parent */
	free(elem);
	return next;
}


void aspect_coverage_traceCoverageElement(
		const char *aspect,
		const char *structName,
		const char **elems,
		size_t count)
{
TraceNode *curNode, *rootNode;
size_t ctr ;

    if (!aspectrace_allowed(ASPECT, "element"))
		return;

	assertion( elems != NULL && elems[0] != NULL,
		FORMAT("aspect_coverage_traceCoverageElement: Empty name chain") );

	{
		const char *rootAttrs[] = { "aspect", aspect, "cs", structName, "name", elems[0], NULL };
		rootNode = curNode = create_TraceNode( "element", rootAttrs, NULL, 0 );
	}

    for (ctr = 1; ctr < count; ctr++) {
        const char *nextName = elems[ ctr ];
        const char *attrs[] = { "name", nextName, NULL };
        curNode->firstChild = create_TraceNode( "element", attrs, NULL, 0 );
        curNode = curNode->firstChild;
    }
    aspect_trace( ASPECT, NULL, rootNode );
}

void aspect_coverage_traceCoverageElement_a(
		const char *aspect,
		const char *structName,
		const char **elems)
{
    aspect_coverage_traceCoverageElement(aspect, structName, elems, list_len((const void**)elems));
}

static TraceNode *convert_CoverageElement_to_TraceNode(aspect_coverage_element *ce)
{
TraceNode *res, **childrenNodes;
const char *attrs[] = { "name", ce->name, NULL };
aspect_coverage_element *curChild;
size_t i, count;

	curChild = ce->firstChild;
	for (count = 0; curChild != NULL; count++)
		curChild = curChild->nextSibling;

	childrenNodes = (TraceNode**)calloc( count, sizeof(TraceNode*) );
	assertion( childrenNodes != 0, FORMAT("convert_CoverageElement_to_TraceNode: No memory") );

	curChild = ce->firstChild;
	for (i = 0; curChild != NULL; i++) {
		aspect_coverage_element *nextChild = curChild->nextSibling;
		childrenNodes[i] = convert_CoverageElement_to_TraceNode( curChild );	/* destroys curChild */
		curChild = nextChild;
	}

	res = create_TraceNode("element", attrs, childrenNodes, count);

	/*** Cleanup ***/

	/* All children are already destroyed by convert_CoverageElement_to_TraceNode().
	 * Do not let function destroy_CoverageElement(ce) access them. */
	ce->firstChild = NULL;

	free( childrenNodes );
	destroy_CoverageElement( ce );

	return res;
}

void aspect_coverage_traceCoverageStructure(
		const char *aspect,
		const char *name,
		aspect_coverage_element **elems,
		size_t count)
{
TraceNode **elemNodes;
size_t i;

	if (!aspectrace_allowed(ASPECT, "structure"))
		return;

	assertion( elems != NULL && elems[0] != NULL,
		FORMAT("aspect_coverage_traceCoverageStructure: Empty coverage structure") );
	assertion( aspect != NULL && name != NULL,
		FORMAT("aspect_coverage_traceCoverageStructure: NULL parameter") );

	/* Count coverage elements */
	if (count > 0)
	{
		elemNodes = (TraceNode **)calloc( count, sizeof(TraceNode*) );
		assertion( elemNodes != 0, FORMAT("aspect_coverage_traceCoverageStructure: No memory") );

		/* Create NULL-terminated array of TraceNode's */
		for (i = 0; i < count; i++) {
			elemNodes[i] = convert_CoverageElement_to_TraceNode( elems[i] );
		}
	}
	else
	{
		elemNodes = NULL;
	}

	// Create and trace message
	{
		const char *rootAttrs[] = { "aspect", aspect, "name", name, NULL };
		aspect_trace( ASPECT, NULL,
			create_TraceNode("structure", rootAttrs, elemNodes, count)
		);
	}

	if (elemNodes != NULL)
	{
		free( elemNodes );
	}
}

void aspect_coverage_traceCoverageStructure_a(
		const char *aspect,
		const char *name,
		aspect_coverage_element **elems)
{
    aspect_coverage_traceCoverageStructure(aspect, name, elems, list_len((const void**)elems));
}

void aspect_coverage_traceCteskCS(CoverageDesc *cs)
{
int i;
TraceNode **children;

	if (!aspectrace_allowed(ASPECT, "structure"))
		return;

	assertion( cs != NULL, FORMAT("aspect_coverage_traceCteskCS: NULL argument") );
	assertion( cs->kind == PRIMARY_COVERAGE,
		FORMAT("aspect_coverage_traceCteskCS: derived coverage structures not supported") );

	children = (TraceNode**)calloc( cs->size, sizeof(TraceNode*) );
	assertion( children != NULL, FORMAT("aspect_coverage_traceCteskCS: No memory") );

	for (i = 0; i < cs->size; i++) {
		const char *attrs[] = { "name", cs->data[i], NULL };
		children[i] = create_TraceNode("element", attrs, NULL, 0);
	}

	{
		const char *rootAttrs[] = { "aspect", "oracle", "name", cs->name, NULL };
		aspect_trace(ASPECT, NULL, create_TraceNode("structure", rootAttrs, children, cs->size));
	}

	free( children );
}

