/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <ctype.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#if defined(__GNUC__) && defined(LIBXML_ENABLED)
    #include <libxml/xmlreader.h>
#endif

#include <atl/object_int.h>
#include <generator.h>
#include <generator/exception.h>
#include <generator/tracer.h>
#include <ts/engine_tsi.h>
#include <ts/register_tsi.h>
#include <ts/system.h>
#include <utils/assertion.h>
#include <ts/cpp_iface.h>

#if defined(__GNUC__)
    #include <pthread.h>
#elif defined(WIN32)
    #include <windows.h>
#else
    #error "unknown platform"
#endif

/******************************************************************************/
/**                               Test Scenario                              **/
/******************************************************************************/

/******************************************************************************/
/**                                Test Engine                               **/
/******************************************************************************/

/******************************************************************************/
/**                       Test Engine / Basic Data Types                     **/
/******************************************************************************/

static void init_NodeId(struct TE_NodeId *node_id, va_list *arg_list)
{
    node_id->id  = va_arg(*arg_list, String*);
    node_id->deg = va_arg(*arg_list, int);

#ifdef NETFSM_ITERATION_CACHE
    // In the beginning, the iteration cache is uninitialized
    node_id->sysdata.cache.index = -1;
    node_id->sysdata.cache.kind  = -1;
    node_id->sysdata.cache.value = NULL;
#endif // NETFSM_ITERATION_CACHE
}

static void copy_NodeId(struct TE_NodeId *src, struct TE_NodeId *dst)
{
#ifdef NETFSM_CLONE_NODES
    // Deep copy of the string identifier
    dst->id  = src->id != NULL ? create_String(toCharArray_String(r(src->id))) : NULL;
#else
    // Only pointer to the string identifier is copied
    dst->id  = src->id != NULL ? clone_Object(r(src->id)) : NULL;
#endif // NETFSM_CLONE_NODES

    dst->deg = src->deg;

#ifdef NETFSM_ITERATION_CACHE
    // Iteration cache is not copied
    dst->sysdata.cache.index = -1;
    dst->sysdata.cache.kind  = -1;
    dst->sysdata.cache.value = NULL;
#endif // NETFSM_ITERATION_CACHE
}

static int compare_NodeId(struct TE_NodeId *lhs, struct TE_NodeId *rhs)
{
    return compare(r(lhs->id), r(rhs->id));
}

static String* to_string_NodeId(struct TE_NodeId *node_id)
{
    return r(node_id->id);
}

static String* to_XML_NodeId(struct TE_NodeId *node_id)
{
    return to_XML_spec("TE_NodeId", to_string_NodeId(node_id));
}

static void destroy_NodeId(struct TE_NodeId *node_id)
{
    destroy(node_id->id);

#ifdef NETFSM_ITERATION_CACHE
    destroy(node_id->sysdata.cache.value);
#endif // NETFSM_ITERATION_CACHE
}

const Type type_TE_NodeId =
{
    "TE_NodeId"
   , sizeof(TE_NodeId)
   , &type_Object
   , (Init)init_NodeId
   , (Copy)copy_NodeId
   , (Compare)compare_NodeId
   , (ToString)to_string_NodeId
   , (ToXML)to_XML_NodeId
   , (CheckInvariant)check_invariant_Default
   , (Enumerate)enumerate_Default
   , (Destroy)destroy_NodeId
};

TE_NodeId* create_NodeId(String *id)
{
    return create(&type_TE_NodeId, id, TE_UNKNOWN_DEGREE);
}

TE_NodeId* create_NodeIdWithDegree(String *id, int degree)
{
    return create(&type_TE_NodeId, id, degree);
}

int hash_NodeId(TE_NodeId *node_id)
{
    int h = hash_String(r(node_id->id));

    destroy(node_id);

    return h;
}

TE_StimulusId create_StimulusId(int index)
{
    TE_StimulusId stimulus_id = { /*.index =*/ index };

    return stimulus_id;
}

int compare_StimulusId(TE_StimulusId lhs, TE_StimulusId rhs)
{
    int res;

    if((res = lhs.index - rhs.index) != 0)
        { return res; }

    return 0;
}

bool equal_StimulusId(TE_StimulusId lhs, TE_StimulusId rhs)
{
    return lhs.index == rhs.index;
}

static void init_ArcId(struct TE_ArcId *arc_id, va_list *arg_list)
{
    arc_id->start_node_id = va_arg(*arg_list, TE_NodeId*);
    arc_id->stimulus_id   = va_arg(*arg_list, TE_StimulusId);
    arc_id->end_node_id   = va_arg(*arg_list, TE_NodeId*);
}

static void copy_ArcId(struct TE_ArcId *src, struct TE_ArcId *dst)
{
    dst->start_node_id = src->start_node_id != NULL ? (TE_NodeId *)clone_Object(r(src->start_node_id)) : NULL;
    dst->end_node_id   = src->end_node_id   != NULL ? (TE_NodeId *)clone_Object(r(src->end_node_id))   : NULL;
    dst->stimulus_id   = src->stimulus_id;
}

static int compare_ArcId(struct TE_ArcId *lhs, struct TE_ArcId *rhs)
{
    int res;

    if((res = lhs->start_node_id != NULL) != (rhs->start_node_id != NULL))
        { return -1; }
    if(res && (res = compare(r(lhs->start_node_id), r(rhs->start_node_id))) != 0)
        { return res; }
    if((res = lhs->end_node_id != NULL) != (rhs->end_node_id != NULL))
        { return -1; }
    if(res && (res = compare(r(lhs->end_node_id), r(rhs->end_node_id))) != 0)
        { return res; }
    if((res = compare_StimulusId(lhs->stimulus_id, rhs->stimulus_id)) != 0)
        { return res; }

    return 0;
}

static String* to_string_ArcId(struct TE_ArcId *arc_id)
{
    String *string = format_String("(%s, %d, %s)",
        NODE_ID_STRING(arc_id->start_node_id),
        arc_id->stimulus_id.index,
        NODE_ID_STRING(arc_id->end_node_id));

    return string;
}

static String* to_XML_ArcId(struct TE_ArcId *arc_id)
{
    return to_XML_spec("TE_ArcId", to_string_ArcId(arc_id));
}

static void destroy_ArcId(struct TE_ArcId *arc_id)
{
    destroy(arc_id->start_node_id);
    destroy(arc_id->end_node_id);
}

const Type type_TE_ArcId =
{
    "TE_ArcId"
   , sizeof(TE_ArcId)
   , &type_Object
   , (Init)init_ArcId
   , (Copy)copy_ArcId
   , (Compare)compare_ArcId
   , (ToString)to_string_ArcId
   , (ToXML)to_XML_ArcId
   , (CheckInvariant)check_invariant_Default
   , (Enumerate)enumerate_Default
   , (Destroy)destroy_ArcId
};

TE_ArcId* create_ArcId(TE_NodeId *start_node_id, TE_StimulusId stimulus_id, TE_NodeId *end_node_id)
{
    return create(&type_TE_ArcId, start_node_id, stimulus_id, end_node_id);        
}                                                          

/******************************************************************************/
/**                       Test Engine / Graph Data Base                      **/
/******************************************************************************/

void init_GraphInfo(TE_GraphInfo *graph_info)
{
    assertion(graph_info != NULL, "init_GraphInfo: graph_info == NULL");
    memset(graph_info, 0, sizeof(TE_GraphInfo));
}

void init_GraphParameters(TE_GraphParameters *graph_params)
{
    assertion(graph_params != NULL, "init_GraphParameters: graph_params == NULL");
    memset(graph_params, 0, sizeof(TE_GraphParameters));
}

static void init_NoSyncMutex  (TE_GraphMutex *mutex) { }
static void lock_NoSyncMutex  (TE_GraphMutex *mutex) { }
static void unlock_NoSyncMutex(TE_GraphMutex *mutex) { }
static void finish_NoSyncMutex(TE_GraphMutex *mutex) { }

TE_GraphMutex TE_NO_SYNC_MUTEX =
{
    /*.init   =*/ init_NoSyncMutex,
    /*.lock   =*/ lock_NoSyncMutex,
    /*.unlock =*/ unlock_NoSyncMutex,
    /*.finish =*/ finish_NoSyncMutex,
    /*.handle =*/ NULL
};

static void init_DefaultMutex(TE_GraphMutex *mutex)
{
#if defined(__GNUC__)
    pthread_mutex_t data = PTHREAD_MUTEX_INITIALIZER;

    mutex->handle = malloc(sizeof(pthread_mutex_t));
    assertion(mutex->handle != NULL, "init_DefaultMutex: out of memory");

    memcpy( mutex->handle, &data, sizeof(pthread_mutex_t) );
#elif defined(WIN32)
    (HANDLE*)mutex->handle = CreateMutex(NULL, FALSE, NULL);
    assertion(mutex->handle != NULL, "init_DefaultMutex: mutex->handle == NULL");
#endif
}

static void lock_DefaultMutex(TE_GraphMutex *mutex)
{
    int status;
#if defined(__GNUC__)
    status = pthread_mutex_lock((pthread_mutex_t*)mutex->handle);
    assertion(status == 0, "lock_DefaultMutex: status != 0");
#elif defined(WIN32)
    status = WaitForSingleObject((HANDLE*)mutex->handle, INFINITE);
    assertion(status == WAIT_OBJECT_0, "lock_DefaultMutex: status != WAIT_OBJECT_0");
#endif
}

static void unlock_DefaultMutex(TE_GraphMutex *mutex)
{
    int status;
#if defined(__GNUC__)
    status = pthread_mutex_unlock((pthread_mutex_t*)mutex->handle);
    assertion(status == 0, "unlock_DefaultMutex: status != 0");
#elif defined(WIN32)
    status = ReleaseMutex((HANDLE*)mutex->handle);
    assertion(status != 0, "unlock_DefaultMutex: status == 0");
#endif
}

static void finish_DefaultMutex(TE_GraphMutex *mutex)
{
#if defined(__GNUC__)
    if(mutex->handle != NULL) { free(mutex->handle); }
#elif defined(WIN32)
    if(mutex->handle != NULL) { CloseHandle((HANDLE*)mutex->handle); }
#endif
}

TE_GraphMutex TE_DEFAULT_MUTEX =
{
    /*.init   =*/ init_DefaultMutex,
    /*.lock   =*/ lock_DefaultMutex,
    /*.unlock =*/ unlock_DefaultMutex,
    /*.finish =*/ finish_DefaultMutex,
    /*.handle =*/ NULL
};

void init_Graph(TE_Graph *graph, TE_UpdateMode mode, TE_GraphMutex *mutex)
{
    assertion(graph != NULL, "init_Graph: graph == NULL");
    assertion(mutex != NULL, "init_Graph: mutex == NULL");

    init_GraphInfo(&graph->info);
    init_GraphParameters(&graph->params);

    graph->status    = TE_DISABLED;
    graph->pause_req = false;
    graph->pause_ack = false;
    graph->mutex     = mutex;
    graph->length    = 0;
    graph->error     = 0;
    graph->fails     = 0;
    graph->updates   = mode == TE_UPDATE_ON ? create_List(&type_TE_ArcId) : NULL;
    graph->data      = NULL;

    graph->mutex->init(graph->mutex);
}

void destroy_Graph(TE_Graph *graph)
{
    assertion(graph != NULL, "destroy_Graph: graph == NULL");

    graph->mutex->finish(graph->mutex);

    if(graph->updates != NULL)
        { destroy(graph->updates); }
}

void lock_Graph(TE_Graph *graph)
{
    graph->mutex->lock(graph->mutex);
}

void unlock_Graph(TE_Graph *graph)
{
    graph->mutex->unlock(graph->mutex);
}

TE_GraphStatus getStatus_Graph(TE_Graph *graph)
{
    TE_GraphStatus status;

    assertion(graph != NULL, "getStatus_Graph: graph == NULL");

    lock_Graph(graph);
    {
        status = graph->status;
    }
    unlock_Graph(graph);

    return status;
}

int getLength_Graph(TE_Graph *graph)
{
    int length;

    assertion(graph != NULL, "getLength_Graph: graph == NULL");

    lock_Graph(graph);
    {
        length = graph->length;
    }
    unlock_Graph(graph);

    return length;
}

int incLength_Graph(TE_Graph *graph, int path_length)
{
    int length;

    assertion(graph != NULL, "incLength_Graph: graph == NULL");

    lock_Graph(graph);
    {
        length = (graph->length += path_length);
    }
    unlock_Graph(graph);

    return length;
}

int getFailures_Graph(TE_Graph *graph)
{
    int fails;

    assertion(graph != NULL, "getFailures_Graph: graph == NULL");

    lock_Graph(graph);
    {
        fails = graph->fails;
    }
    unlock_Graph(graph);

    return fails;
}

int incFailures_Graph(TE_Graph *graph)
{
    int fails;

    assertion(graph != NULL, "incFailures_Graph: graph == NULL");

    lock_Graph(graph);
    {
        fails = ++graph->fails;
    }
    unlock_Graph(graph);

    return fails;
}

int getError_Graph(TE_Graph *graph)
{
    int error;

    assertion(graph != NULL, "getError_Graph: graph == NULL");

    lock_Graph(graph);
    {
        error = graph->error;
    }
    unlock_Graph(graph);

    return error;
}

void sleep_Graph(TE_Graph *graph)
{
    assertion(graph != NULL, "sleep_Graph: graph == NULL");

    for(;;) 
    {
        lock_Graph(graph);
        {
            graph->pause_ack = graph->pause_req;
        }
        unlock_Graph(graph);

        if(!graph->pause_ack) { break; }

        sleep_msec(1);
    }
}

void printProgress_Graph(TE_Graph *graph)
{
    assertion(graph != NULL, "printProgress_Graph: graph == NULL");

    lock_Graph(graph);
    {
        int nodes_finished  = graph->info.nodes_finished;
	int nodes_traversed = graph->info.nodes_traversed;
	int arcs_traversed  = graph->info.arcs_traversed;
	int arcs_known      = graph->info.arcs_known;

        printf("[SYSTEM] TRAVERSAL_PROGRESS: states %d/%d, transitions %d/%d, length %d [%.1f%%]\n",
            nodes_finished, nodes_traversed, arcs_traversed, arcs_known,
	    graph->length, (100.0*arcs_traversed)/arcs_known);
    }
    unlock_Graph(graph);
}

/******************************************************************************/
/**                          Test Engine / Wrappers                          **/
/******************************************************************************/

#define DECLARE_EXCEPTION(name) \
    int name##_exc = 0; \
    const char *name##_msg = NULL;

#define THROW_EXCEPTION_IF(cond, name, label, msg) \
    if(cond) { name##_exc = 1; name##_msg = msg; goto label; }

#define HANDLE_EXCEPTION(name) \
    assertion(name##_exc == 0, name##_msg);

const char* getName_TestEngine(TestEngine *engine)
{
    assertion(engine != NULL, "getName_TestEngine: engine == NULL");

    return engine->name;
}

void init_TestEngine(TestEngine *engine, TE_Graph *graph, TestScenario *test, int argc, char **argv)
{
    DECLARE_EXCEPTION(e);

    assertion(engine != NULL, "init_TestEngine: engine == NULL");
    assertion(graph  != NULL, "init_TestEngine: graph == NULL");
    assertion(test   != NULL, "init_TestEngine: test == NULL");

    assertion(argc >= 0,                 "init_TestEngine: argc < 0");
    assertion(argc == 0 || argv != NULL, "init_TestEngine: argc > 0 && argv == NULL");

    assertion(engine->init      != NULL, "init_TestEngine: engine->init == NULL");
    assertion(engine->config    != NULL, "init_TestEngine: engine->config == NULL");
    assertion(engine->start     != NULL, "init_TestEngine: engine->start == NULL");
    assertion(engine->update    != NULL, "init_TestEngine: engine->update == NULL");
    assertion(engine->getPath   != NULL, "init_TestEngine: engine->getPath == NULL");
    assertion(engine->reachNode != NULL, "init_TestEngine: engine->reachNode == NULL");
    assertion(engine->getError  != NULL, "init_TestEngine: engine->getError == NULL");
    assertion(engine->finish    != NULL, "init_TestEngine: engine->finish == NULL");

    lock_Graph(graph);
    {
        THROW_EXCEPTION_IF(graph->status != TE_DISABLED, e,
            unlock_graph_db, "init_TestEngine: incorrect status");
    
        graph->status = TE_ENABLED;
        engine->init(graph, argc, argv);
    
        // Initialize the arc executor
        engine->sysdata.executor = (ArcSymbolExecutor*)malloc(sizeof(ArcSymbolExecutor));

        THROW_EXCEPTION_IF(engine->sysdata.executor == NULL, e,
            unlock_graph_db, "init_TestEngine: executor == NULL");

        engine->sysdata.executor->actions                       = test->actions;
        // FIXME Update definition of executor, enable usage of functors in saveModelState etc.
        engine->sysdata.executor->model_state.saveModelState    = test->saveModelState;
        engine->sysdata.executor->model_state.restoreModelState = test->restoreModelState;
        engine->sysdata.executor->model_state.isStationaryState = test->isStationaryState;
        engine->sysdata.executor->model_state.observeState      = test->observeState;
    }
unlock_graph_db:
    unlock_Graph(graph);
    HANDLE_EXCEPTION(e);
}

static void setNodeDegree_TestEngine(TestEngine *engine, TE_NodeId *node_id)
{
    ArcSymbol symbol;
    ScenarioFunctionID *actions = engine->sysdata.executor->actions;

    node_id->deg = 0;

    for(symbol.kind = 0; actions[symbol.kind] != NULL; symbol.kind++)
    {
        symbol.value = action_init(actions[symbol.kind]);

        for(;;)
        {
            if(!action_next(actions[symbol.kind], r(symbol.value)))
               { break; }

            node_id->deg++;
        }

        destroy(symbol.value); 
    }

#ifdef DEBUG_NETFSM
    printf("setNodeDegree_TestEngine: node_id=%s, degree=%d\n", NODE_ID_STRING(node_id), node_id->deg);
#endif
}

static void expand_TestEngine(TestEngine *engine, TE_Graph *graph, TE_NodeId *node_id)
{
    TE_ArcId *arc_id;
    TE_StimulusId stimulus_id;

    assertion(node_id->deg >= 0, "expand_TestEngine: node_id->deg < 0");

#ifdef DEBUG_NETFSM
    printf("expand_TestEngine: node_id=%s\n", NODE_ID_STRING(node_id));
#endif

    for(stimulus_id.index = 0; stimulus_id.index < node_id->deg; stimulus_id.index++)
    {
        // End nodes of untraversed arcs are NULL
        arc_id = create_ArcId(r(node_id), stimulus_id, NULL);
        engine->update(graph, TE_LOCAL, arc_id, NULL);
        destroy(arc_id);
    }
}

void start_TestEngine(TestEngine *engine, TE_Graph *graph, TE_NodeId *node_id)
{
    DECLARE_EXCEPTION(e);

    assertion(engine  != NULL, "start_TestEngine: engine == NULL");
    assertion(graph   != NULL, "start_TestEngine: graph == NULL");
    assertion(node_id != NULL, "start_TestEngine: node_id == NULL");

#ifdef DEBUG_NETFSM
    printf("start_TestEngine: node_id=%s\n", NODE_ID_STRING(node_id));
#endif

    lock_Graph(graph);
    {
        THROW_EXCEPTION_IF(graph->status == TE_DISABLED, e,
            unlock_graph_db, "start_TestEngine: incorrect status: TE_DISABLED");
        THROW_EXCEPTION_IF(graph->status == TE_STARTED, e,
            unlock_graph_db, "start_TestEngine: incorrect status: TE_STARTED");

        // The database might have been already destroyed by the synchronizer
        if(graph->status == TE_ENABLED)
        {
            if(engine->start(graph, node_id) == TE_NEW_NODE)
            {
                setNodeDegree_TestEngine(engine, node_id);
                expand_TestEngine(engine, graph, node_id);
            }

            graph->status = TE_STARTED;
        }
    }
unlock_graph_db:
    unlock_Graph(graph);
    HANDLE_EXCEPTION(e);
}

TE_UpdateStatus updateLocal_TestEngine(TestEngine *engine, TE_Graph *graph, TE_ArcId *arc_id, TE_NodeId *new_end_node_id)
{
    TE_UpdateStatus status = TE_OLD_ALL;

    DECLARE_EXCEPTION(e);

    assertion(engine                != NULL, "updateLocal_TestEngine: engine == NULL");
    assertion(graph                 != NULL, "updateLocal_TestEngine: graph == NULL");
    assertion(arc_id                != NULL, "updateLocal_TestEngine: arc_id == NULL");
    assertion(arc_id->start_node_id != NULL, "updateLocal_TestEngine: start_node_id == NULL");
    assertion(new_end_node_id       != NULL, "updateLocal_TestEngine: new_end_node_id == NULL");

    lock_Graph(graph);
    {
        THROW_EXCEPTION_IF(graph->status == TE_DISABLED, e,
            unlock_graph_db, "updateLocal_TestEngine: incorrect status: TE_DISABLED");
        THROW_EXCEPTION_IF(graph->status == TE_ENABLED, e,
            unlock_graph_db, "updateLocal_TestEngine: incorrect status: TE_ENABLED");
    
        // The database might have been already destroyed by the synchronizer
        if(graph->status == TE_STARTED)
        {
            status = engine->update(graph, TE_LOCAL, arc_id, new_end_node_id);
    
            if((status & TE_NEW_NODE) != 0)
            {
                setNodeDegree_TestEngine(engine, new_end_node_id);
                expand_TestEngine(engine, graph, new_end_node_id);
            }
    
            if((status & TE_NEW_ARC) != 0)
            {
                if(graph->updates != NULL)
                {
                    // Loops are not sent to the synchronizer (their execution usually does not require much time)
                    if(!equals(r(arc_id->start_node_id), r(new_end_node_id)))
                    {
                        // This code is to avoid memory leaks
                        if(arc_id->end_node_id != NULL)
                            { destroy(arc_id->end_node_id); }

                        // A clone of the arc_id is added to the list
                        arc_id->end_node_id = r(new_end_node_id);
                        append_List(r(graph->updates), (TE_ArcId *)clone_Object(r(arc_id)));
                    }
                }
            }
        }
    }
unlock_graph_db:
    unlock_Graph(graph);
    HANDLE_EXCEPTION(e);

    return status;
}

TE_UpdateStatus updateRemote_TestEngine(TestEngine *engine, TE_Graph *graph, TE_ArcId *arc_id)
{
    TE_UpdateStatus status = TE_OLD_ALL;

    DECLARE_EXCEPTION(e);

    assertion(engine                != NULL, "updateRemote_TestEngine: engine == NULL");
    assertion(graph                 != NULL, "updateRemote_TestEngine: graph == NULL");
    assertion(arc_id                != NULL, "updateRemote_TestEngine: arc_id == NULL");
    assertion(arc_id->start_node_id != NULL, "updateRemote_TestEngine: start_node_id == NULL");
    assertion(arc_id->end_node_id   != NULL, "updateRemote_TestEngine: end_node_id == NULL");

    lock_Graph(graph);
    {
        THROW_EXCEPTION_IF(graph->status == TE_DISABLED, e,
            unlock_graph_db, "updateRemote_TestEngine: incorrect status: TE_DISABLED");
        THROW_EXCEPTION_IF(graph->status == TE_ENABLED, e,
            unlock_graph_db, "updateRemote_TestEngine: incorrect status: TE_ENABLED");
    
        // The database might have been already destroyed by the traverser
        if(graph->status == TE_STARTED)
        {
            status = engine->update(graph, TE_REMOTE, arc_id, NULL);

            if((status & TE_NEW_NODE) != 0)
            {
                // Node degree should have been already set
                expand_TestEngine(engine, graph, arc_id->end_node_id);
            }

            // Destroy the clone
            destroy(arc_id);
        }
    }
unlock_graph_db:
    unlock_Graph(graph);
    HANDLE_EXCEPTION(e);

    return status;
}

void updateRemoteAll_TestEngine(TestEngine *engine, TE_Graph *graph, TE_ArcList *arc_list)
{
    int i;

    DECLARE_EXCEPTION(e);

    assertion(engine   != NULL, "updateRemoteAll_TestEngine: engine == NULL");
    assertion(graph    != NULL, "updateRemoteAll_TestEngine: graph == NULL");
    assertion(arc_list != NULL, "updateRemoteAll_TestEngine: arc_list == NULL");

    lock_Graph(graph);
    {
        THROW_EXCEPTION_IF(graph->status == TE_DISABLED, e,
            unlock_graph_db, "updateRemoteAll_TestEngine: incorrect status: TE_DISABLED");
        THROW_EXCEPTION_IF(graph->status == TE_ENABLED, e,
            unlock_graph_db, "updateRemoteAll_TestEngine: incorrect status: TE_ENABLED");

        // The database might have been already destroyed by the traversal
        if(graph->status == TE_STARTED)
        {    
            for(i = 0; i < size_List(r(arc_list)); i++)
            {
                TE_UpdateStatus status = TE_OLD_ALL;
                TE_ArcId *arc_id = (TE_ArcId*)get_List(r(arc_list), i);
        
                THROW_EXCEPTION_IF(arc_id->start_node_id == NULL, e,
                    unlock_graph_db, "updateRemoteAll_TestEngine: start_node_id == NULL");
        
                status = engine->update(graph, TE_REMOTE, arc_id, NULL);

                if((status & TE_NEW_NODE) != 0)
                {
                    // Node degree should have been already set
                    expand_TestEngine(engine, graph, arc_id->end_node_id);
                }

                // Destroy the clone
                destroy(arc_id);
            }
        }
    }
unlock_graph_db:
    unlock_Graph(graph);
    HANDLE_EXCEPTION(e);
}

TE_ArcList* getUpdates_TestEngine(TestEngine *engine, TE_Graph *graph)
{
    TE_ArcList *updates = NULL;
    DECLARE_EXCEPTION(e);

    assertion(engine != NULL, "getUpdates_TestEngine: engine == NULL");
    assertion(graph  != NULL, "getUpdates_TestEngine: graph == NULL");

    lock_Graph(graph);
    {
        THROW_EXCEPTION_IF(graph->updates == NULL, e,
            unlock_graph_db, "getUpdates_TestEngine: graph->updates == NULL");
        THROW_EXCEPTION_IF(graph->status == TE_DISABLED, e,
            unlock_graph_db, "getUpdates_TestEngine: incorrect status: TE_DISABLED");
        THROW_EXCEPTION_IF(graph->status == TE_ENABLED, e,
            unlock_graph_db, "getUpdates_TestEngine: incorrect status: TE_ENABLED");
    
        if(graph->updates != NULL)
        {
            updates = graph->updates;
            graph->updates = create_List(&type_TE_ArcId);
        }
    }
unlock_graph_db:
    unlock_Graph(graph);
    HANDLE_EXCEPTION(e);

    // The list should be destroyed by the synchronizer
    return updates;
}

TE_ArcList* getPath_TestEngine(TestEngine *engine, TE_Graph *graph, TE_NodeId *node_id)
{
    TE_ArcList *path = NULL;

    DECLARE_EXCEPTION(e);

    assertion(engine  != NULL, "getPath_TestEngine: engine == NULL");
    assertion(graph   != NULL, "getPath_TestEngine: graph == NULL");
    assertion(node_id != NULL, "getPath_TestEngine: node_id == NULL");

    lock_Graph(graph);
    {
        THROW_EXCEPTION_IF(graph->status == TE_DISABLED, e,
            unlock_graph_db, "getPath_TestEngine: incorrect status: TE_DISABLED");
        THROW_EXCEPTION_IF(graph->status == TE_ENABLED, e,
            unlock_graph_db, "getPath_TestEngine: incorrect status: TE_ENABLED");
    
        // The database might have been already destroyed by the synchronizer
        if(graph->status == TE_STARTED)
            { path = engine->getPath(graph, node_id); }
    }
unlock_graph_db:
    unlock_Graph(graph);
    HANDLE_EXCEPTION(e);

    return path;
}

TE_ArcList* reachNode_TestEngine(TestEngine *engine, TE_Graph *graph, TE_NodeId *node_id, TE_ArcId *arc_id)
{
    TE_ArcList *path = NULL;

    DECLARE_EXCEPTION(e);

    assertion(engine  != NULL, "reachNode_TestEngine: engine == NULL");
    assertion(graph   != NULL, "reachNode_TestEngine: graph == NULL");
    assertion(node_id != NULL, "reachNode_TestEngine: node_id == NULL");

    lock_Graph(graph);
    {
        THROW_EXCEPTION_IF(graph->status == TE_DISABLED, e,
            unlock_graph_db, "reachNode_TestEngine: incorrect status: TE_DISABLED");
        THROW_EXCEPTION_IF(graph->status == TE_ENABLED, e,
            unlock_graph_db, "reachNode_TestEngine: incorrect status: TE_ENABLED");
    
        // The database might have been already destroyed by the synchronizer
        if(graph->status == TE_STARTED)
            { path = engine->reachNode(graph, node_id, arc_id); }
    }
unlock_graph_db:
    unlock_Graph(graph);
    HANDLE_EXCEPTION(e);

    return path;
}

String* getError_TestEngine(TestEngine *engine, TE_Graph *graph)
{
    String *error;

    DECLARE_EXCEPTION(e);

    assertion(engine != NULL, "getError_TestEngine: engine == NULL");
    assertion(graph  != NULL, "getError_TestEngine: graph == NULL");

    lock_Graph(graph);
    {
        THROW_EXCEPTION_IF(graph->status == TE_DISABLED, e,
            unlock_graph_db, "getError_TestEngine: incorrect status");

        error = engine->getError(graph);
    }
unlock_graph_db:
    unlock_Graph(graph);
    HANDLE_EXCEPTION(e);

    return error;
}

void getInfo_TestEngine(TestEngine *engine, TE_Graph *graph, TE_GraphInfo *info)
{
    TE_ArcList *path = NULL;

    DECLARE_EXCEPTION(e);

    assertion(engine != NULL, "getInfo_TestEngine: engine == NULL");
    assertion(graph  != NULL, "getInfo_TestEngine: graph == NULL");
    assertion(info   != NULL, "getInfo_TestEngine: info == NULL");

    lock_Graph(graph);
    {
        THROW_EXCEPTION_IF(graph->status == TE_DISABLED, e,
            unlock_graph_db, "getInfo_TestEngine: incorrect status");

        *info = graph->info;
    }
unlock_graph_db:
    unlock_Graph(graph);
    HANDLE_EXCEPTION(e);
}

void finish_TestEngine(TestEngine *engine, TE_Graph *graph)
{
    DECLARE_EXCEPTION(e);

    assertion(engine != NULL, "finish_TestEngine: engine == NULL");
    assertion(graph  != NULL, "finish_TestEngine: graph == NULL");

    lock_Graph(graph);
    {
        THROW_EXCEPTION_IF(graph->status == TE_DISABLED, e,
            unlock_graph_db, "finish_TestEngine: incorrect status");

        // The database might have been already destroyed by some other component
        if(graph->status != TE_DESTROYED)
        {
            engine->finish(graph);

            // Destroy the arc executor
            free(engine->sysdata.executor);

            graph->status = TE_DESTROYED;
        }
    }
unlock_graph_db:
    unlock_Graph(graph);
    HANDLE_EXCEPTION(e);
}

bool toFile_TestEngine(TestEngine *engine, TE_Graph *graph, const char *filename)
{
    bool res = true;

    DECLARE_EXCEPTION(e);

    assertion(engine != NULL, "toFile_TestEngine: engine == NULL");
    assertion(graph  != NULL, "toFile_TestEngine: graph == NULL");

    lock_Graph(graph);
    {
        THROW_EXCEPTION_IF(graph->status == TE_DISABLED , e,
            unlock_graph_db, "toFile_TestEngine: incorrect status: TE_DISABLED");
        THROW_EXCEPTION_IF(graph->status == TE_DESTROYED , e,
            unlock_graph_db, "toFile_TestEngine: incorrect status: TE_DESTROYED");
    
        res = engine->toFile(graph, filename);
    }
unlock_graph_db:
    unlock_Graph(graph);
    HANDLE_EXCEPTION(e);

    return res;
}

bool fromFile_Default(TestEngine *engine, TE_Graph *graph, const char *filename)
{
#if defined(__GNUC__) && defined(LIBXML_ENABLED)
    int status;
    bool start_node = true;
    xmlTextReaderPtr reader;
    TE_StimulusId stimulus_id;
    TE_NodeId *start_node_id = NULL;
    TE_NodeId *end_node_id = NULL;

    LIBXML_TEST_VERSION

    assertion(engine   != NULL, "fromFile_Default: engine == NULL");
    assertion(graph    != NULL, "fromFile_Default: graph == NULL");
    assertion(filename != NULL, "fromFile_Default: filename == NULL");

    reader = xmlReaderForFile(filename, NULL, 0);

    if(reader == NULL)
        { return false; }

    while((status = xmlTextReaderRead(reader)) == 1)
    {
        const xmlChar *name  = xmlTextReaderConstName(reader);
        const xmlChar *value = xmlTextReaderConstValue(reader);

        int type = xmlTextReaderNodeType(reader);

        if(!strcmp(name, "start_node"))
        {
            if(type == XML_READER_TYPE_ELEMENT)
            {
                assertion(start_node_id == NULL, "fromFile_Default: unexpected 'start_node'");

                if((status = xmlTextReaderRead(reader)) != 1)
                    { break; }

                value = xmlTextReaderConstValue(reader);
                start_node_id = strcmp(value, "unknown") ? create_NodeId(create_String(value)) : NULL;
            }
        }
        else if(!strcmp(name, "end_node"))
        {
            if(type == XML_READER_TYPE_ELEMENT)
            {
                assertion(end_node_id == NULL, "fromFile_Default: unexcepted 'end_node'");

                if((status = xmlTextReaderRead(reader)) != 1)
                    { break; }

                value = xmlTextReaderConstValue(reader);
                end_node_id = strcmp(value, "unknown") ? create_NodeId(create_String(value)) : NULL;
            }
        }
        else if(!strcmp(name, "index"))
        {
            if(type == XML_READER_TYPE_ELEMENT)
            {
                if((status = xmlTextReaderRead(reader)) != 1)
                    { break; }

                value = xmlTextReaderConstValue(reader);
                stimulus_id.index = atoi(value);
            }
        }
        else if(!strcmp(name, "transition"))
        {
            if(type == XML_READER_TYPE_END_ELEMENT)
            {
                TE_ArcId *arc_id = create_ArcId(start_node_id, stimulus_id, end_node_id);

                if(start_node)
                {
                    engine->start(graph, start_node_id);
                    start_node = false;
                }

                engine->update(graph, TE_LOCAL, arc_id, end_node_id);

                start_node_id = end_node_id = NULL;
            }
        }
    }

    xmlFreeTextReader(reader);

    assertion(status == 0, "fromFile_Default: incorrect format");
#endif
    return true;
}

bool fromFile_TestEngine(TestEngine *engine, TE_Graph *graph, const char *file)
{
    bool res = true;

    DECLARE_EXCEPTION(e);

    assertion(engine != NULL, "fromFile_TestEngine: engine == NULL");
    assertion(graph  != NULL, "fromFile_TestEngine: graph == NULL");

    lock_Graph(graph);
    {
        THROW_EXCEPTION_IF(graph->status != TE_ENABLED, e,
            unlock_graph_db, "fromFile_TestEngine: incorrect status");
    
        if(engine->fromFile != NULL)
            { res = engine->fromFile(graph, file); }
        else
            { res = fromFile_Default(engine, graph, file); }
    }
unlock_graph_db:
    unlock_Graph(graph);
    HANDLE_EXCEPTION(e);

    return res;
}

static void traceEnd_TestEngine(TestEngine *engine, TE_Graph *graph, time_t start_time, time_t end_time)
{
    String *msg;

    DECLARE_EXCEPTION(e);

    assertion(engine != NULL, "traceEnd_TestEngine: engine == NULL");
    assertion(graph  != NULL, "traceEnd_TestEngine: graph == NULL");

    lock_Graph(graph);
    {
        THROW_EXCEPTION_IF(graph->status == TE_DISABLED, e,
            unlock_graph_db, "traceEnd_TestEngine: incorrect status");

        if(graph->status == TE_ENABLED)
        {
            traceExceptionInfo_TestEngine("SCENARIO_NOT_STARTED", "Test scenario has not been started");
            goto unlock_graph_db;
        }

        if(graph->status == TE_STARTED)
        {
            uint32_t test_time = (uint32_t)difftime(end_time, start_time);
            Map *summary = create_Map(&type_String, &type_String);

            String *node_key = create_String("States");
            String *node_val = format_String("%d/%d (local/total)",
                graph->info.nodes_traversed_locally, graph->info.nodes_traversed);

            String *arcs_key = create_String("Transitions");
            String *arcs_val = format_String("%d/%d (local/total)",
                graph->info.arcs_traversed_locally, graph->info.arcs_traversed);

            String *loop_key = create_String("Selfloops");
            String *loop_val = format_String("%d/%d (loops/total)",
                graph->info.loops_traversed, graph->info.arcs_traversed);

            String *ndet_key = create_String("Nondeterminism");
            String *ndet_val = format_String("%d/%d (transitions)",
                graph->info.arcs_nondeterministic, graph->info.arcs_traversed);

            String *trav_key = create_String("Traversal");
            String *trav_val = format_String("%d (transitions), %d (seconds)",
                graph->length, test_time);

            put_Map(r(summary), node_key, node_val);
            put_Map(r(summary), arcs_key, arcs_val);
            put_Map(r(summary), loop_key, loop_val);
            put_Map(r(summary), ndet_key, ndet_val);
            put_Map(r(summary), trav_key, trav_val);

            traceTestSummary_TestEngine(summary);
            destroy(summary);
        }

        if(graph->error == 0 && graph->status == TE_STARTED)
        {
            traceSystemInfo_TestEngine("SCENARIO_COMPLETED", "Test scenario has been successfully completed");
        }
        else if(graph->error == 0 && graph->status == TE_DESTROYED)
        {
            traceExceptionInfo_TestEngine("SCENARIO_ABORTED", "Test scenario has been aborted");
        }
        else if((msg = engine->getError(graph)) == NULL)
        {
            msg = format_String("Unknown error has occured: errno=%d", graph->error);

            traceExceptionInfo_TestEngine("UNKNOWN_ERROR", toCharArray_String(r(msg)));
            destroy(msg);
        }
        else
        {
            traceExceptionInfo_TestEngine("UNCONNECTED_GRAPH", toCharArray_String(r(msg)));
            destroy(msg);
        }

        flushTrace_TestEngine();
    }
unlock_graph_db:
    unlock_Graph(graph);
    HANDLE_EXCEPTION(e);
}

void suspend_TestEngine(TestEngine *engine, TE_Graph *graph)
{
    bool pause_ack;

    assertion(graph != NULL, "suspend_TestEngine: graph == NULL");

    for(;;)
    {
        lock_Graph(graph);
        {
            graph->pause_req = true;
            pause_ack = graph->pause_ack;
        }
        unlock_Graph(graph);

        if(pause_ack) { break; }

        sleep_msec(1);
    }
}

void resume_TestEngine(TestEngine *engine, TE_Graph *graph)
{
    assertion(graph != NULL, "resume_TestEngine: graph == NULL");

    lock_Graph(graph);
    {
        graph->pause_req = false;
    }
    unlock_Graph(graph);
}

bool isSuspended_TestEngine(TestEngine *engine, TE_Graph *graph)
{
    bool suspended;

    assertion(graph != NULL, "isSuspended_TestEngine: graph == NULL");

    lock_Graph(graph);
    {
        suspended = graph->pause_req;
    }
    unlock_Graph(graph);

    return suspended;
}

/******************************************************************************/
/**                        Test Engine / Attributes                          **/
/******************************************************************************/

String* getAttrValue_TestEngine(TestEngine *engine, TE_Graph *graph, const char *attr_name)
{
    TE_GraphInfo graph_info;

    assertion(engine    != NULL, "getAttrValue_TestEngine: engine == NULL");
    assertion(graph     != NULL, "getAttrValue_TestEngine: graph == NULL");
    assertion(attr_name != NULL, "getAttrValue_TestEngine: attr_name == NULL");

    if(!strcmp(attr_name, TE_ATTR_CPU_ID))
    {
        getInfo_TestEngine(engine, graph, &graph_info);
        return format_String("%03d", graph_info.cpu_id);
    }
    if(!strcmp(attr_name, TE_ATTR_TRAVERSED_NODES))
    {
        getInfo_TestEngine(engine, graph, &graph_info);
        return format_String("%d", graph_info.nodes_traversed);
    }
    if(!strcmp(attr_name, TE_ATTR_LOCALLY_TRAVERSED_NODES))
    {
        getInfo_TestEngine(engine, graph, &graph_info);
        return format_String("%d", graph_info.nodes_traversed_locally);
    }
    if(!strcmp(attr_name, TE_ATTR_FINISHED_NODES))
    {
        getInfo_TestEngine(engine, graph, &graph_info);
        return format_String("%d", graph_info.nodes_finished);
    }
    if(!strcmp(attr_name, TE_ATTR_KNOWN_ARCS))
    {
        getInfo_TestEngine(engine, graph, &graph_info);
        return format_String("%d", graph_info.arcs_known);
    }
    if(!strcmp(attr_name, TE_ATTR_NONDETERM_ARCS))
    {
        getInfo_TestEngine(engine, graph, &graph_info);
        return format_String("%d", graph_info.arcs_nondeterministic);
    }
    if(!strcmp(attr_name, TE_ATTR_TRAVERSED_ARCS))
    {
        getInfo_TestEngine(engine, graph, &graph_info);
        return format_String("%d", graph_info.arcs_traversed);
    }
    if(!strcmp(attr_name, TE_ATTR_LOCALLY_TRAVERSED_ARCS))
    {
        getInfo_TestEngine(engine, graph, &graph_info);
        return format_String("%d", graph_info.arcs_traversed_locally);
    }
    if(!strcmp(attr_name, TE_ATTR_TRAVERSED_LOOPS))
    {
        getInfo_TestEngine(engine, graph, &graph_info);
        return format_String("%d", graph_info.loops_traversed);
    }
    if(!strcmp(attr_name, TE_ATTR_STATUS))
    {
        switch(getStatus_Graph(graph))
        {
        case TE_DISABLED  : { return create_String(TE_ATTR_STATUS_DISABLED); }
        case TE_ENABLED   : { return create_String(TE_ATTR_STATUS_ENABLED); }
        case TE_STARTED   : { return create_String(TE_ATTR_STATUS_STARTED); }
        case TE_DESTROYED :
            {
                getInfo_TestEngine(engine, graph, &graph_info);

                if(getError_Graph(graph) != 0 || getFailures_Graph(graph) > 0)
                    { return create_String(TE_ATTR_STATUS_FAILED); }

                if(graph_info.arcs_traversed < graph_info.arcs_known)
                    { return create_String(TE_ATTR_STATUS_ABORTED); }

                return create_String(TE_ATTR_STATUS_FINISHED);
            }
        default: { assertion(false, "getAttrValue_TestEngine: unknown status"); }
        }
    }
    if(!strcmp(attr_name, TE_ATTR_FAILURES))
    {
        return format_String("%d", getFailures_Graph(graph));
    }
    if(!strcmp(attr_name, TE_ATTR_ERROR))
    {
        return getError_TestEngine(engine, graph);
    }

    return NULL;
}

Map* getAllAttrValues_TestEngine(TestEngine *engine, TE_Graph *graph)
{
    TE_GraphInfo graph_info;
    String *status_value;
    Map *all_attr_values = create_Map(&type_String, &type_String);

    assertion(engine != NULL, "getAllAttrValues_TestEngine: engine == NULL");
    assertion(graph  != NULL, "getAllAttrValues_TestEngine: graph == NULL");

    getInfo_TestEngine(engine, graph, &graph_info);

    switch(getStatus_Graph(graph))
    {
    case TE_DISABLED  : { status_value = create_String(TE_ATTR_STATUS_DISABLED); break; }
    case TE_ENABLED   : { status_value = create_String(TE_ATTR_STATUS_ENABLED);  break; }
    case TE_STARTED   : { status_value = create_String(TE_ATTR_STATUS_STARTED);  break; }
    case TE_DESTROYED :
        {
            if(getError_Graph(graph) != 0 || getFailures_Graph(graph) > 0)
                { status_value = create_String(TE_ATTR_STATUS_FAILED); }
            else if(graph_info.arcs_traversed < graph_info.arcs_known)
                { status_value = create_String(TE_ATTR_STATUS_ABORTED); }
            else
                { status_value = create_String(TE_ATTR_STATUS_FINISHED); }
            break;
        }
    default: { assertion(false, "getAllAttrValues_TestEngine: unknown status"); }
    }

    put_Map(r(all_attr_values), create_String(TE_ATTR_CPU_ID),
        format_String("%03d", graph_info.cpu_id));
    put_Map(r(all_attr_values), create_String(TE_ATTR_TRAVERSED_NODES),
        format_String("%d", graph_info.nodes_traversed));
    put_Map(r(all_attr_values), create_String(TE_ATTR_LOCALLY_TRAVERSED_NODES),
        format_String("%d", graph_info.nodes_traversed_locally));
    put_Map(r(all_attr_values), create_String(TE_ATTR_FINISHED_NODES),
        format_String("%d", graph_info.nodes_finished));
    put_Map(r(all_attr_values), create_String(TE_ATTR_KNOWN_ARCS),
        format_String("%d", graph_info.arcs_known));
    put_Map(r(all_attr_values), create_String(TE_ATTR_NONDETERM_ARCS),
        format_String("%d", graph_info.arcs_nondeterministic));
    put_Map(r(all_attr_values), create_String(TE_ATTR_TRAVERSED_ARCS),
        format_String("%d", graph_info.arcs_traversed));
    put_Map(r(all_attr_values), create_String(TE_ATTR_LOCALLY_TRAVERSED_ARCS),
        format_String("%d", graph_info.arcs_traversed_locally));
	put_Map(r(all_attr_values), create_String(TE_ATTR_TRAVERSED_LOOPS),
		format_String("%d", graph_info.loops_traversed));
    put_Map(r(all_attr_values), create_String(TE_ATTR_STATUS),
        status_value);
    put_Map(r(all_attr_values), create_String(TE_ATTR_FAILURES),
        format_String("%d", getFailures_Graph(graph)));
    put_Map(r(all_attr_values), create_String(TE_ATTR_ERROR),
        getError_TestEngine(engine, graph));

    return all_attr_values;
}

void setAttrValue_TestEngine(TestEngine *engine, TE_Graph *graph, String *attr_name, String *attr_value)
{
    DECLARE_EXCEPTION(e);

    assertion(engine     != NULL, "setAttrValue_TestEngine: engine == NULL");
    assertion(graph      != NULL, "setAttrValue_TestEngine: graph == NULL");
    assertion(attr_name  != NULL, "setAttrValue_TestEngine: attr_name == NULL");
    assertion(attr_value != NULL, "setAttrValue_TestEngine: attr_value == NULL");

    lock_Graph(graph);
    {
        THROW_EXCEPTION_IF(graph->status == TE_DISABLED, e,
            unlock_graph_db, "setAttrValue_TestEngine: incorrect status");

        if(graph->status != TE_DESTROYED)
            { engine->config(graph, toCharArray_String(r(attr_name)), toCharArray_String(r(attr_value))); }

    }
unlock_graph_db:
    unlock_Graph(graph);
    HANDLE_EXCEPTION(e);    
}

void setAllAttrValues_TestEngine(TestEngine *engine, TE_Graph *graph, Map *all_attrs)
{
    int i;
    String *attr_name;
    String *attr_value;
        
    DECLARE_EXCEPTION(e);

    assertion(engine    != NULL, "setAllAttrValues_TestEngine: engine == NULL");
    assertion(graph     != NULL, "setAllAttrValues_TestEngine: graph == NULL");
    assertion(all_attrs != NULL, "setAllAttrValues_TestEngine: all_attrs == NULL");

    lock_Graph(graph);
    {
        THROW_EXCEPTION_IF(graph->status == TE_DISABLED, e,
            unlock_graph_db, "setAllAttrValues_TestEngine: incorrect status");

        if(graph->status != TE_DESTROYED)
        {
            for(i = 0; i < size_Map(all_attrs); i++)
            {
                attr_name  = key_Map(r(all_attrs), i);
                attr_value = get_Map(r(all_attrs), r(attr_name));

                engine->config(graph, toCharArray_String(r(attr_name)), toCharArray_String(r(attr_value)));

                destroy(attr_name);
                destroy(attr_value);
            }
        }

    }
unlock_graph_db:
    unlock_Graph(graph);
    HANDLE_EXCEPTION(e);    
}

/******************************************************************************/
/**                       Test Engine / Main Function                        **/
/******************************************************************************/

/**
 * Returns the arc symbol (iteration context) corresponding to a given stimulus.
 * Performs a series of consequent iterations from the beginning to the given one.
 */
static ArcSymbol getArcSymbol_TestEngine(TestEngine *engine, TE_ArcId *arc_id)
{
    ArcSymbol symbol;
    int index = arc_id->stimulus_id.index;
    ScenarioFunctionID *actions = engine->sysdata.executor->actions;

    assertion(index >= 0, "getArcSymbol_TestEngine: arc_id->stimulus_id.index < 0");

    symbol.kind  = 0;
    symbol.value = NULL;

#ifdef NETFSM_ITERATION_CACHE
    // call() makes one iteration before invoking an arc symbol
    int next_index = arc_id->stimulus_id.index + 1;
    int node_range = arc_id->start_node_id->deg;

    TE_Iteration *cache = &(arc_id->start_node_id->sysdata.cache);

    if(cache->value != NULL)
    {
        if(cache->index <= index)
        {
            // Restore an iteration from the cache
            index       -= cache->index;
            symbol.kind  = cache->kind;
            symbol.value = cache->value;
        }
	else
	{
            // Invalidate the cache
	    destroy(cache->value);
	    cache->value = NULL;
	}
    }
#endif // NETFSM_ITERATION_CACHE

    for(; actions[symbol.kind] != NULL; symbol.kind++)
    {
#ifdef NETFSM_ITERATION_CACHE
        if(symbol.value == NULL)
#endif // NETFSM_ITERATION_CACHE
            { symbol.value = action_init(actions[symbol.kind]); }

        for(;;)
        {
            if(index == 0)
            { 
                if(action_next(actions[symbol.kind], clone_Object(r(symbol.value))))
		{
#ifdef NETFSM_ITERATION_CACHE
                    // Store the iteration into the cache
                    cache->index = next_index;
                    cache->kind  = symbol.kind;
                    cache->value = (next_index >= node_range ? NULL : r(symbol.value));
#endif // NETFSM_ITERATION_CACHE
		    return symbol;
                }
                else
                    { break; }
            }
            else
            {
                if(action_next(actions[symbol.kind], r(symbol.value)))
                    { index--; }
                else
                    { break; }
            }
        }

        destroy(symbol.value);
	symbol.value = NULL;
    }

    assertion(false, "getArcSymbol_TestEngine: arc_id->stimulus_id.index is out of range");

    return symbol;    
}

static bool executeArcSymbol_TestEngine(TestEngine *engine, TE_ArcId *arc_id)
{
    bool old;
    bool verdict;
    ArcSymbol symbol;
    String *tran_id;

    ScenarioFunctionID *actions = engine->sysdata.executor->actions;

    assertion(arc_id != NULL, "executeArcSymbol_TestEngine: arc_id == NULL");

    tran_id = format_String("(%s, %d)",
        NODE_ID_STRING(arc_id->start_node_id), arc_id->stimulus_id.index);

    // Get the test action
    symbol = getArcSymbol_TestEngine(engine, arc_id);

    old = traceTransitionStart_TestEngine(toCharArray_String(r(tran_id)),
        action_name(actions[symbol.kind]), symbol.value, arc_id->end_node_id == NULL);

    // Apply the test action
    verdict = action_invoke(actions[symbol.kind], r(symbol.value));
    destroy(symbol.value);

    verdict = onFrameEnd(&engine->sysdata.executor->model_state, verdict);

    traceTransitionEnd_TestEngine(toCharArray_String(r(tran_id)),
        action_name(actions[symbol.kind]), old);

    destroy(tran_id);

    return verdict;
}

#define GET_NODE(test) \
    (scenario_get_state((test)))

#define GET_NODE_ID(test, node) \
    (create_NodeId(scenario_get_state_id((test), r(node))))

#define GET_NODE_STRING(node) \
    (toString(r(node)))

static void traverse_TestEngine(TestEngine *engine, TE_Graph *graph, TestScenario *test)
{
    int i;
    int print_length = 0;
    int print_delta  = 0;
    time_t start_time, end_time;

    Object    *current_node        = GET_NODE(test);
    TE_NodeId *current_node_id     = GET_NODE_ID(test, current_node);
    String    *current_node_string = GET_NODE_STRING(current_node);

    // Get start time
    time(&start_time);

    start_TestEngine(engine, graph, current_node_id);

    while(getStatus_Graph(graph) == TE_STARTED)
    {
        TE_ArcList *path;

        // Suspend traversal if it is needed
        sleep_Graph(graph);

        // Test has been completed
        if((path = getPath_TestEngine(engine, graph, current_node_id)) == NULL)
        {
            // Get finish time
            time(&end_time);

            // Print traversal progress
            if(graph->params.print_progress)
                { printProgress_Graph(graph); }

            traceEnd_TestEngine(engine, graph, start_time, end_time);
            goto finish_traversal;
        }

        // Walk along the path
        for(i = 0; i < size_List(r(path)); i++)
        {
            TE_ArcId *arc_id = get_List(r(path), i);

            traceState_TestEngine(NODE_ID_STRING(current_node_id), toCharArray_String(r(current_node_string)));

            if(!executeArcSymbol_TestEngine(engine, arc_id))
            {
                int failures = incFailures_Graph(graph);

                if(tsNumberOfErrorsToFinish > 0 && failures >= tsNumberOfErrorsToFinish)
                    { destroy(arc_id); destroy(path); goto finish_traversal; }
            }

            destroy(current_node);
            destroy(current_node_id);
            destroy(current_node_string);

            current_node        = GET_NODE(test);
            current_node_id     = GET_NODE_ID(test, current_node);
            current_node_string = GET_NODE_STRING(current_node);

            if(arc_id->end_node_id == NULL)
            {
                updateLocal_TestEngine(engine, graph, arc_id, current_node_id);
            }
            else if(!equals(r(arc_id->end_node_id), r(current_node_id)))
            {
                i++;

                traceNondeterminism_TestEngine
                (
                    NODE_ID_STRING(arc_id->start_node_id),
                    arc_id->stimulus_id.index,
                    NODE_ID_STRING(arc_id->end_node_id),
                    NODE_ID_STRING(current_node_id)
                );

                updateLocal_TestEngine(engine, graph, arc_id, current_node_id);

                destroy(arc_id);

                break;
            }

            destroy(arc_id);
        }

        // Increment the traversal length
        incLength_Graph(graph, i);

        // Print traversal progress
        if(graph->params.print_progress)
        {
	    int length = getLength_Graph(graph);

            if(length >= print_length + print_delta)
            {
                printProgress_Graph(graph);

                print_length = length;
		print_delta += (print_delta >> 1) + 1;

		if(print_delta > 12345)
		    { print_delta = 12345; }
            }
        }

        // Destroy the path
        destroy(path);
    }

finish_traversal:
    traceState_TestEngine(NODE_ID_STRING(current_node_id), toCharArray_String(r(current_node_string)));

    destroy(current_node);
    destroy(current_node_id);
    destroy(current_node_string);
}

static bool checkConsistency_TestScenario(TestScenario *test)
{
    if(areDeferredReactionsEnabled())
    {
        if(test->isStationaryState == NULL)
        {
            traceExceptionInfo_TestEngine(
                "INCONSISTENT_SCENARIO",
                "is_stationary field should be initialized if deferred reactions are enabled");
            return false;
        }
        if(test->saveModelState == NULL)
        {
            traceExceptionInfo_TestEngine(
                "INCONSISTENT_SCENARIO",
                "save_model_state field should be initialized if deferred reactions are enabled");
            return false;
        }
        if(test->restoreModelState == NULL)
        {
            traceExceptionInfo_TestEngine(
                "INCONSISTENT_SCENARIO",
                "restore_model_state field should be initialized if deferred reactions are enabled");
            return false;
        }
    }

    return true;
}

static void traceProperties(void)
{
    char buff[64];

    // Deferred Reactions Enabled
    traceProperty_TestEngine("Deferred Reactions Enabled", areDeferredReactionsEnabled() ? "true" : "false");

    if(areDeferredReactionsEnabled())
    {
        // WTime
        if(getWTimeMSec() % 1000 == 0)
            { sprintf(buff, "%d secs", (int)getWTime()); }
        else
            { sprintf(buff, "%d millisecs", (int)getWTimeMSec()); }

        traceProperty_TestEngine("WTime", buff);

        // Is Find First Series Only
        traceProperty_TestEngine("Find First Series Only", isFindFirstSeriesOnly() ? "true" : "false" );
    }

    // Time Model
    switch(getTSTimeModel())
    {
    case NotUseTSTime:
        traceProperty_TestEngine("Time Model", "not used");
        break;
    case LinearTSTime:
        traceProperty_TestEngine("Time Model", "linear");
        break;
	case DistributedTSTime :
        traceProperty_TestEngine("Time Model", "distributed");
        break;
	default:
        assertion(false, "Unexpected Time Model");
    }

    // Finish Mode
    switch (getFinishMode())
    {
    case UNTIL_ERROR :
        if(tsNumberOfErrorsToFinish == 1)
        {
            traceProperty_TestEngine("Finish Mode", "until an error");
        }
        else
        {
            sprintf(buff, "until %d errors", tsNumberOfErrorsToFinish);
            traceProperty_TestEngine("Finish Mode", buff);
        }
        break;
    case UNTIL_END :
        traceProperty_TestEngine("Finish Mode", "until end");
        break;
    default :
        assertion( false, "Unexpected Finish Mode" );
    }
}

static void processGraphParameters_TestEngine(TE_Graph *graph, int argc, char **argv)
{
    int i;
    assertion(graph != NULL, "processGraphParameters_TestEngine: graph == NULL");

    for(i = 0; i < argc; i++) 
    {
        if(!strcmp(argv[i], "--print-progress"))
	    { graph->params.print_progress = true; }
    }
}

bool main_TestEngine(TestEngine *engine, TE_Graph *graph, TestScenario *test, int argc, char **argv)
{
    bool tracing_was_started_here = false;
    bool was_accidental_transition_traced;
    bool verdict, verdict_init;
    String *cpu_id;
    TestSystemParameters *pars;

    // Change assertion handler
    atAssertion(netfsm_default_assertion_handler);

    assertion(engine != NULL, "main_TestEngine: engine == NULL");
    assertion(graph  != NULL, "main_TestEngine: graph == NULL");
    assertion(test   != NULL, "main_TestEngine: test == NULL");

    // Normalize test data
    if(test->getState == NULL)
        { scenario_set_default_get_state(test); }
    if(test->getStateId == NULL)
        { scenario_set_default_get_state_id(test); }

    // Init DR mode and time model
    if((test->isStationaryState != NULL) &&
       (test->saveModelState    != NULL) &&
       (test->restoreModelState != NULL))
    {
        setDeferredReactionsMode(true);
        setTSTimeModel(LinearTSTime);
    }
    else
    {
        setDeferredReactionsMode(false);
        setTSTimeModel(NotUseTSTime);
    }

    // Process netfsm parameters (argc and argv are not changed)
    processGraphParameters_TestEngine(graph, argc, argv);

    // Process standard parameters
    cpu_id = getAttrValue_TestEngine(engine, graph, TE_ATTR_CPU_ID);
    assertion(cpu_id != NULL, "main_TestEngine: cpu_id == NULL");

    pars = processStandardParameters(test->name, toCharArray_String(r(cpu_id)), test->actions, argc, argv);
    assertion(pars != NULL, "main_TestEngine: pars == NULL");

    // Process trace parameters
    processTraceParameters_TestEngine(test->name, toCharArray_String(r(cpu_id)), argc, argv);

    destroy(cpu_id);

    // Start tracing if it has not been started yet
    tracing_was_started_here = startTrace_TestEngine();
    
    START_EXCEPTION_PROTECTED_SECTION(netfsm_default_exception_handler);

    setCurrentTestScenarioName(test->name);

    traceScenarioStart_TestEngine(test->name, pars->argc, pars->argv);

    // Scenario initialization
    verdict = true;
    if(test->init != NULL)
    {
        // Trace pseudo transition start
        traceState_TestEngine("start", "start");
        was_accidental_transition_traced = traceTransitionStart_TestEngine("initialize", "initialize", NULL, true);
    
        // Call init function
        clearOracleVerdict();

        verdict_init = scenario_init(test, pars->argc, pars->argv);
        verdict = verdict_init && getOracleVerdict();

        // Frame end
        verdict = onFrameEnd(&engine->sysdata.executor->model_state, verdict);
    
        // Trace initialization failed message
        if(!verdict_init)
        {
            traceExceptionInfo_TestEngine("SCENARIO_INITIALIZATION_FAILED",
                "Scenario initialization has been failed");

            testVerdict = TD_faulted;
        }
    
        // Trace pseudo transition end
        traceTransitionEnd_TestEngine("initialize", "initialize", was_accidental_transition_traced);
    }

    // FSM traversal
    if(verdict)
    {
        if(checkConsistency_TestScenario(test))
        {
            traceProperties();

            traverse_TestEngine(engine, graph, test);
 
            // Evaluate the historical test verdict
            if(graph->error != 0)
            {
                verdict = false;
                testVerdict = TD_faulted;
            }
            else if(graph->fails > 0)
            {
                verdict = false;
                testVerdict = TD_bad_verdict;
            }
            else
            {
                testVerdict = TD_ok;
            }
        }
 
        if(test->finish != NULL)
        {
            // Trace pseudo transition start
            was_accidental_transition_traced = traceTransitionStart_TestEngine("finalize", "finalize", NULL, true);
     
            // Call finish function
            clearOracleVerdict();
            scenario_finish(test);
            verdict = verdict && getOracleVerdict();
 
            // Frame end
            verdict = onFrameEnd(&engine->sysdata.executor->model_state, verdict);
     
            // Trace pseudo transition end
            traceTransitionEnd_TestEngine("finalize", "finalize", was_accidental_transition_traced);
            traceState_TestEngine("end", "end");
        }
    }

    traceScenarioEnd_TestEngine();
    setCurrentTestScenarioName(NULL);

    // Deallocate resources
    finish_TestEngine(engine, graph);

    END_EXCEPTION_PROTECTED_SECTION();

    // Stop tracing if it was started here
    endTrace_TestEngine(tracing_was_started_here);

    // Disable standard parameters
    disableStandardParameters(pars);

    return verdict;
}

/******************************************************************************/
/**                       Test Engine / User Functions                       **/
/******************************************************************************/

static TestEngine *__current_engine = NULL;
static TE_Graph   *__current_graph  = NULL;

void set_TestEngine(TestEngine *engine, TE_Graph *graph)
{
    assertion(engine != NULL, "set_TestEngine: engine == NULL");
    assertion(graph  != NULL, "set_TestEngine: graph == NULL");

    __current_engine = engine;
    __current_graph  = graph;
}

bool start_TestScenario(int argc, char **argv, TestScenario *test)
{
    return main_TestEngine(__current_engine, __current_graph, test, argc, argv);
}

int localmain(TestEngine *engine, TestScenario *test, int argc, char **argv)
{
    int verdict;
    TE_Graph graph;

    init_Graph(&graph, TE_UPDATE_OFF, &TE_NO_SYNC_MUTEX);
    init_TestEngine(engine, &graph, test, argc, argv);
    verdict = main_TestEngine(engine, &graph, test, argc, argv);
    destroy_Graph(&graph);

    return !verdict;
}
